# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""This module contains prompt tuning through PEFT"""
# Standard
from datetime import datetime
from typing import Any, Callable, Dict, Iterable, List, Optional, Tuple, Union
import gc
import json
import os

# Third Party
from accelerate import Accelerator
from peft import (
    MultitaskPromptTuningConfig,
    PeftConfig,
    PeftModel,
    PeftType,
    PromptTuningConfig,
    TaskType,
    get_peft_model,
)
from torch.optim import AdamW
from torch.utils.data import DataLoader
from tqdm import tqdm
from transformers import AutoModelForCausalLM, default_data_collator
from transformers.models.auto.tokenization_auto import AutoTokenizer
from transformers.optimization import get_linear_schedule_with_warmup
import numpy as np
import torch
import transformers

# First Party
from caikit.core.data_model import DataStream
from caikit.core.exceptions import error_handler
from caikit.core.modules import ModuleBase, ModuleConfig, ModuleSaver, module
from caikit.interfaces.nlp.data_model import (
    ClassificationTrainRecord,
    GeneratedTextResult,
    GeneratedTextStreamResult,
    TokenizationResults,
)
from caikit.interfaces.nlp.tasks import TextGenerationTask, TokenizationTask
import alog

# Local
from ...data_model import (
    ExponentialDecayLengthPenalty,
    GenerationTrainRecord,
    PromptOutputModelType,
    TuningConfig,
)
from ...resources.pretrained_model import (
    HFAutoCausalLM,
    HFAutoSeq2SeqLM,
    PretrainedModelBase,
)
from ...toolkit.data_stream_wrapper import SimpleIterableStreamWrapper
from ...toolkit.data_type_utils import get_torch_dtype, str_to_torch_dtype
from ...toolkit.task_specific_utils import convert_to_generation_record
from ...toolkit.text_generation.model_run_utils import (
    GENERATE_FUNCTION_ARGS,
    generate_text_func,
    generate_text_func_stream,
)
from ...toolkit.trainer_utils import validate_training_data
from ...toolkit.verbalizer_utils import render_verbalizer
from .peft_config import TuningType, get_peft_config, resolve_base_model

log = alog.use_channel("PEFT_PROMPT")
error = error_handler.get(log)

TRAINING_LOSS_LOG_FILENAME = "training_logs.jsonl"

# TODO: try to refactor this into a smaller module
# pylint: disable=too-many-lines,too-many-instance-attributes
@module(
    id="6655831b-960a-4dc5-8df4-867026e2cd41",
    name="Peft generation",
    version="0.1.0",
    tasks=[TextGenerationTask, TokenizationTask],
)
class PeftPromptTuning(ModuleBase):

    _DETECT_DEVICE = "__DETECT__"
    _ENCODER_KEY = PromptOutputModelType.ENCODER
    _DECODER_KEY = PromptOutputModelType.DECODER
    _ADAPTER_NAME = "default"

    tuning_type_to_huggingface = {
        TuningType.PROMPT_TUNING: PeftType.PROMPT_TUNING,
        TuningType.MULTITASK_PROMPT_TUNING: PeftType.MULTITASK_PROMPT_TUNING,
        # TuningType.MULTITASK_PREFIX_TUNING: PeftType.MULTITASK_PREFIX_TUNING,
        # TuningType.P_TUNING: PeftType.P_TUNING,
        # TuningType.PREFIX_TUNING: PeftType.PREFIX_TUNING,
        # TuningType.LORA: PeftType.LORA,
    }

    RANDOM_SEED = 73
    supported_resources = [HFAutoCausalLM, HFAutoSeq2SeqLM]

    ################################ Constructor / Destructor #####################################

    def __init__(
        self,
        tokenizer: AutoTokenizer,
        model: PeftModel,
        base_model_config: Dict[str, Any],
        base_model_name: str,
        verbalizer: str,
        task_type: str,
        tuning_type: TuningType,
        output_model_types: List[PromptOutputModelType],
        training_metadata: Union[Dict[str, Any], None] = None,
    ):
        super().__init__()
        # Put the PEFT model into evaluation mode for all future calls
        model.eval()
        self._collate_fn = self._get_collate_fn(tokenizer, task_type)
        self.model = model
        self.tokenizer = tokenizer
        self.base_model_name = base_model_name
        self._base_model_config = base_model_config
        self.eos_token_id = self.tokenizer.encode(self.tokenizer.eos_token)[-1]
        self.verbalizer = verbalizer
        self.task_type = task_type
        self.tuning_type = tuning_type
        self.output_model_types = output_model_types
        self.training_metadata = (
            training_metadata if training_metadata is not None else {}
        )

    # pylint: disable=duplicate-code
    def __del__(self):
        del self.model
        del self.tokenizer
        gc.collect()
        try:
            torch.cuda.empty_cache()
        except AttributeError:
            pass

    ################################## API functions #############################################

    @TextGenerationTask.taskmethod()
    def run(
        self,
        text: str,
        max_new_tokens: Optional[int] = 20,
        min_new_tokens: Optional[int] = 0,
        truncate_input_tokens: Optional[int] = 0,
        decoding_method: Optional[str] = "GREEDY",
        top_k: Optional[int] = None,
        top_p: Optional[float] = None,
        typical_p: Optional[float] = None,
        temperature: Optional[float] = None,
        repetition_penalty: Optional[float] = None,
        max_time: Optional[float] = None,
        exponential_decay_length_penalty: Optional[
            Union[Tuple[int, float], ExponentialDecayLengthPenalty]
        ] = None,
        stop_sequences: Optional[List[str]] = None,
        seed: Optional[np.uint64] = None,
        preserve_input_text: bool = True,
    ) -> GeneratedTextResult:
        f"""
        Run the full text generation model.
        Args:
            {GENERATE_FUNCTION_ARGS}
            preserve_input_text: bool
                Applicable only to Causal LLMs.
                Whether or not the source string should be contained in the generated output,
                e.g., as a prefix. Default True. (Source string will appear as prefix)
        Returns:
            GeneratedTextResult
                Generated text result produced by PEFT / Transformers.
        """

        verbalized_text = render_verbalizer(self.verbalizer, {"input": text})

        return generate_text_func(
            self.model,
            self.tokenizer,
            self.PRODUCER_ID,
            self.tokenizer.eos_token,
            verbalized_text,
            max_new_tokens=max_new_tokens,
            min_new_tokens=min_new_tokens,
            truncate_input_tokens=truncate_input_tokens,
            decoding_method=decoding_method,
            top_k=top_k,
            top_p=top_p,
            typical_p=typical_p,
            temperature=temperature,
            seed=seed,
            repetition_penalty=repetition_penalty,
            max_time=max_time,
            exponential_decay_length_penalty=exponential_decay_length_penalty,
            stop_sequences=stop_sequences,
            preserve_input_text=preserve_input_text,
            task_type=self.task_type,
        )

    # NOTE: We need to disable wip decorator here otherwise we get issues in
    # proto generation for streaming. We are keeping it commented out for now,
    # to essentially document that this streaming function is WIP.
    # @wip_decorator.work_in_progress(
    #     category=wip_decorator.WipCategory.WIP, action=wip_decorator.Action.WARNING
    # )
    @TextGenerationTask.taskmethod(output_streaming=True)
    def run_stream_out(
        self,
        text: str,
        max_new_tokens=20,
        min_new_tokens=0,
        truncate_input_tokens: Optional[int] = 0,
        decoding_method: Optional[str] = "GREEDY",
        top_k: Optional[int] = None,
        top_p: Optional[float] = None,
        typical_p: Optional[float] = None,
        temperature: Optional[float] = None,
        repetition_penalty: Optional[float] = None,
        max_time: Optional[float] = None,
        exponential_decay_length_penalty: Optional[
            Union[Tuple[int, float], ExponentialDecayLengthPenalty]
        ] = None,
        stop_sequences: Optional[List[str]] = None,
        seed: Optional[np.uint64] = None,
    ) -> Iterable[GeneratedTextStreamResult]:
        f"""Run the text generation model with output streaming

        NOTE: This implementation is marked as WIP since the API for
        HuggingFace streamer classes at time of implementation is still
        under development and may change.
        Ref. https://huggingface.co/docs/transformers/v4.30.0/generation_strategies#streaming

        Args:
            {GENERATE_FUNCTION_ARGS}

        Returns:
            Iterable[GeneratedTextStreamResult]
        """

        # Apply the verbalizer to our text string
        verbalized_text = render_verbalizer(self.verbalizer, {"input": text})

        return generate_text_func_stream(
            self.model,
            self.tokenizer,
            self.PRODUCER_ID,
            self.tokenizer.eos_token,
            verbalized_text,
            max_new_tokens=max_new_tokens,
            min_new_tokens=min_new_tokens,
            truncate_input_tokens=truncate_input_tokens,
            decoding_method=decoding_method,
            top_k=top_k,
            top_p=top_p,
            typical_p=typical_p,
            temperature=temperature,
            seed=seed,
            repetition_penalty=repetition_penalty,
            max_time=max_time,
            exponential_decay_length_penalty=exponential_decay_length_penalty,
            stop_sequences=stop_sequences,
        )

    @TokenizationTask.taskmethod()
    def run_tokenizer(
        self,
        text: str,
    ) -> TokenizationResults:
        """Run tokenization task against the model

        Args:
           text: str
                Text to tokenize
        Returns:
            TokenizationResults
                The token count
        """
        raise NotImplementedError("Tokenization not implemented for local")

    @classmethod
    def train(
        cls,
        base_model: str,  # TODO: Union[str, PretrainedModelBase]
        train_stream: Union[
            DataStream[GenerationTrainRecord],
            DataStream[ClassificationTrainRecord],
        ],
        tuning_config: TuningConfig,
        val_stream: Optional[
            Union[
                DataStream[GenerationTrainRecord],
                DataStream[ClassificationTrainRecord],
            ]
        ] = None,  # TODO: Optional[DataStream[GenerationTrainRecord]]
        device: Optional[str] = _DETECT_DEVICE,  # TODO: Union[int, str]
        tuning_type: Optional[str] = "PROMPT_TUNING",  # TODO: Union[str, TuningType]
        num_epochs: Optional[int] = 20,
        learning_rate: Optional[float] = 0.3,
        verbalizer: Optional[str] = "{{input}}",
        batch_size: Optional[int] = 8,
        max_source_length: Optional[int] = 256,
        max_target_length: Optional[int] = 128,
        accumulate_steps: Optional[int] = 32,
        torch_dtype: Optional[str] = None,  # TODO: Optional[Union[torch.dtype, str]]
        silence_progress_bars: Optional[bool] = True,
        seed: int = RANDOM_SEED,
        **kwargs,
    ) -> "PeftPromptTuning":
        """Run prompt tuning (vanilla or MPT) through PEFT on a CausalLM or Seq2seq model
        to refine a text generation model.

        Args:
            base_model:  Union[str, caikit_nlp.resources.pretrained_model.base.PretrainedModelBase]
                Base resource model used for underlying generation.
            train_stream: DataStream[GenerationTrainRecord] or DataStream[ClassificationTrainRecord]
                Data to be used for training the prompt vectors of the generation model.
            tuning_config: TuningConfig
                Additional model tuning configurations to be considered for prompt vector
                initialization and training behavior.
            val_stream: Optional[DataStream[GenerationTrainRecord]
                           or DataStream[ClassificationTrainRecord]]
                Data to be used for validation throughout the train process or None.
            device: str
                Device to be used for training the model. Default: cls._DETECT_DEVICE, which
                will fall back to "cuda" if available, else None.
            tuning_type: str
                Type of Peft Tuning config which we would like to build.
            num_epochs: int
                Number of epochs to tune the prompt vectors. Default: 20.
            learning_rate: float
                Learning rate to be used while tuning prompt vectors. Default: 1e-3.
            verbalizer: str
                Verbalizer template to be used for formatting data at train and inference time.
                This template may use brackets to indicate where fields from the data model
                TrainGenerationRecord must be rendered. Default: "{{input}}", i.e., the raw text.
            batch_size: int
                Batch sized to be used for training / evaluation data. Default: 8.
            max_source_length: int
                Max length of input sequences being considered. Default: 256.
            max_target_length: int
                Max length of target sequences being predicted. Default: 128.
            accumulate_steps: int
                Number of steps to use for gradient accumulation. Default: 1.
            torch_dtype: str
                TODO: Optional[Union[torch.dtype, str]]
                Data type to use for training/inference of the underlying text generation model.
                If no value is provided, we pull from torch_dtype in config. If an in memory
                resource is provided which does not match the specified data type, the model
                underpinning the resource will be converted in place to the correct torch dtype.
            silence_progress_bars: bool
                Silences TQDM progress bars at train time. Default: True.
            seed: int
                Integer to be used as random seed for training.
        Returns:
            PeftPromptTuning
                Instance of this class with tuned prompt vectors.
        """
        error.value_check(
            "<NLP46653367E>", len(train_stream) > 0, "train_stream cannot be empty"
        )

        # Configure random seed
        transformers.set_seed(seed)
        # NOTE: Following can be uncommented to allow full determinism
        # but it can have impact on performance.
        # transformers.enable_full_determinism(seed)

        # HACK - These things can't be passed through the train API currently

        metric = kwargs.get("metric")

        base_model = resolve_base_model(base_model, cls, torch_dtype)
        base_model_name = base_model._model_name
        task_type, output_model_types, peft_config, tuning_type = get_peft_config(
            tuning_type,
            tuning_config,
            base_model,
            cls,
            torch_dtype,
            verbalizer,
        )

        # Check if data is within limit allowed for this module and model
        validate_training_data(
            train_stream,
            base_model_name,
            cls.MODULE_ID,
        )

        # Coerce the passed model into a resource; if we have one, this is a noop
        # TODO: When splitting up this mono-module, use the configured resource
        #   type of the concrete class to bootstrap
        torch_dtype = get_torch_dtype(torch_dtype)

        train_stream = train_stream.map(convert_to_generation_record)
        if val_stream:
            error.value_check(
                "<NLP63201425E>", len(val_stream) > 0, "val_stream cannot be empty"
            )

            val_stream = val_stream.map(convert_to_generation_record)

        # Convert our datastreams -> data loaders by disguising them as PyTorch iterable datasets
        train_dataloader, val_dataloader = cls.create_dataloaders_from_stream(
            base_model=base_model,
            task_type=task_type,
            train_stream=train_stream,
            verbalizer=verbalizer,
            validation_stream=val_stream or None,
            batch_size=batch_size,
            max_source_length=max_source_length,
            max_target_length=max_target_length,
        )

        log.debug("Peft config [%s]", peft_config)
        # FIXME: Should only do following line for causal LM (and bloomz?) - check that is the case
        if isinstance(base_model, HFAutoCausalLM):
            base_model.model.config.d_model = 1024

        peft_model = get_peft_model(base_model.model, peft_config)

        # Convert our Peft model (not just the underlying
        # transformers model) to the right underlying type.
        device = cls._get_device(device)
        cls.convert_peft_model_to_type(device, peft_model, torch_dtype)

        training_loss_tracker = cls._execute_train_loop(
            peft_model,
            num_epochs,
            train_dataloader,
            device,
            eval_dataloader=val_dataloader,
            metric=metric,
            learning_rate=learning_rate,
            tokenizer=base_model.tokenizer,
            accumulate_steps=accumulate_steps,
            silence_progress_bars=silence_progress_bars,
            torch_dtype=torch_dtype,
        )

        # Get config of the base model
        base_model_config = base_model.get_config()

        # Remove _name_or_path field as a model can be
        # saved in different location but still same
        del base_model_config["_name_or_path"]
        error.value_check(
            "<NLP07232147E>",
            "_name_or_path" not in base_model_config,
            "_name_or_path needs to be removed from config!",
        )

        # Wrap up the trained model in a class instance
        return cls(
            tokenizer=base_model.tokenizer,
            model=peft_model,
            base_model_config=base_model_config,
            base_model_name=base_model_name,
            verbalizer=verbalizer,
            task_type=task_type,
            tuning_type=tuning_type,
            output_model_types=output_model_types,
            training_metadata=training_loss_tracker,
            # TODO: Export other training params to model as well
        )

    def save(self, model_path: str, save_base_model: bool = False):
        """Save prompt vector and optionally base model in target path

        Args:
            model_path: str
                Path to store model artifact(s)
            save_base_model: bool
                Save base model along with the prompts in the model_path provided.
                Default: False
        """
        module_saver = ModuleSaver(
            self,
            model_path=model_path,
        )

        # NOTE: In case we want optionally allow saving of the base model with the prompts
        # we can use the `base_model.save` method as its a resource that
        # implements its own save method
        prompt_dict = self.get_exportable_prompt_vectors(
            self.model, self.tuning_type, self.output_model_types
        )
        assert prompt_dict, "Failed to export encoder and/or decoder prompts"
        with module_saver:
            config_options = {
                "base_model_config": self._base_model_config,
                "base_model_name": self.base_model_name,
                "eos_token": self.tokenizer.eos_token,
                "has_base_model": save_base_model,
                "verbalizer": self.verbalizer,
                "tuning_type": self.tuning_type.name,
                "task_type": str(self.task_type),
                # Grab the torch property for the dtype so that we can rebuild from a str.
                "trained_torch_dtype": str(self.model.dtype).rsplit(".", maxsplit=1)[
                    -1
                ],
                "output_model_types": json.dumps(
                    [output_type.name for output_type in self.output_model_types]
                ),
            }
            # NOTE: These file names correspond to expected file names in TGIS.
            key_file_pairs = [
                [PeftPromptTuning._ENCODER_KEY.name, "encoder.pt"],
                [PeftPromptTuning._DECODER_KEY.name, "decoder.pt"],
            ]
            for prompt_key, prompt_bin in key_file_pairs:
                prompt_save_path = os.path.realpath(
                    os.path.join(model_path, prompt_bin)
                )

                # Prompt vector (encoder or decoder) not found; set config to empty and continue
                if prompt_dict[prompt_key] is None:
                    config_options[prompt_key] = ""
                    continue

                config_options[prompt_key] = prompt_bin
                log.debug3("Saving prompt %s to: %s", prompt_key, prompt_save_path)
                with alog.ContextTimer(log.debug3, "Done saving prompt in: "):
                    torch.save(prompt_dict[prompt_key], prompt_save_path)
                assert os.path.isfile(
                    prompt_save_path
                ), f"Prompt was not successfully saved to {prompt_save_path}"
            if save_base_model:
                b_model_rel_path, b_model_abs_path = module_saver.add_dir(
                    self.base_model_name
                )
                self.tokenizer.save_pretrained(os.path.join(b_model_abs_path))
                self.model.save_pretrained(os.path.join(b_model_abs_path))

                config_options["full_model_path"] = b_model_rel_path
                config_options["tokenizer_path"] = b_model_rel_path

            training_loss_filename = TRAINING_LOSS_LOG_FILENAME

            config_options.update({"training_logs": training_loss_filename})
            # We are currently only saving logs containing loss in jsonl format
            if "loss" in self.training_metadata:
                loss_log_lines = self.training_metadata.get("loss")
                error.type_check("<NLP60269855E>", list, loss_log_lines=loss_log_lines)
                with open(
                    os.path.join(model_path, training_loss_filename),
                    "w",
                    encoding="utf-8",
                ) as f:
                    for loss_log in loss_log_lines:
                        loss_log = {"name": "loss", "data": loss_log}
                        json.dump(loss_log, f)
                        f.write("\n")

            module_saver.update_config(config_options)

    @classmethod
    def load(
        cls,
        model_path: str,
        torch_dtype: str = None,
        device: str = _DETECT_DEVICE,  # TODO: Union[int, str]
    ) -> "PeftPromptTuning":
        """Load a PEFT prompt tuning model. This method will currently fail if the original
        model was not saved with the arg value save_base_model=True.

        Args:
            model_path: str
                Path to the model to be loaded.
            torch_dtype: str
                Torch data type to be used when loading the model.

        Returns:
            PeftPromptTuning
                Instance of this class built from the on disk model.
        """
        # TODO: Fix this to only allow prompt vector execution
        config = ModuleConfig.load(os.path.abspath(model_path))
        if torch_dtype is not None:
            torch_dtype = str_to_torch_dtype(torch_dtype)
        else:
            torch_dtype = str_to_torch_dtype(config.trained_torch_dtype)
        if config.has_base_model:
            # TODO: Implement logic for resource loading
            device = cls._get_device(device)
            model_config = os.path.join(model_path, config.full_model_path)
            peft_config = PeftConfig.from_pretrained(model_config)
            if peft_config.task_type == "CAUSAL_LM":
                # get the transformers Causal LM model
                base_model = AutoModelForCausalLM.from_pretrained(
                    peft_config.base_model_name_or_path
                )
                # get the PEFT causal LM model
                model = PeftModel.from_pretrained(base_model, model_config)
                cls.convert_peft_model_to_type(device, model, torch_dtype)
            else:
                # TODO: Handle other model types
                error(
                    "<NLP84249238E>",
                    NotImplementedError("Only export of causal LM models is supported"),
                )
            tokenizer = AutoTokenizer.from_pretrained(
                os.path.join(model_path, config.tokenizer_path)
            )
        else:
            # TODO: Can we make this to be a warning and just
            # work with prompt vectors if base model is not provided
            error("<NLP97275192E>", ValueError("base_model not provided."))

        output_model_types = [
            PromptOutputModelType(output_type)
            for output_type in json.loads(config.output_model_types)
        ]

        return cls(
            tokenizer=tokenizer,
            model=model,
            base_model_config=config.base_model_config,
            base_model_name=config.base_model_name,
            verbalizer=config.verbalizer,
            task_type=config.task_type,
            tuning_type=TuningType(config.tuning_type),
            output_model_types=output_model_types,
        )

    ################################## Public Functions ###########################################

    @classmethod
    def get_exportable_prompt_vectors(
        cls,
        model: PeftModel,
        tuning_type: TuningType,
        output_model_types: List[PromptOutputModelType],
    ) -> Dict[str, torch.Tensor]:
        """Grab the prompt vectors off of the model and return a tuple of encoder / decoder
        export vectors.

        Args:
            model: PeftModel
                Model whose prompt vector(s) we want to export.
            tuning_type: TuningType
                Tuning type used to build this model.
            output_model_types: List[PromptOutputModelType]
                Output model types prompt type (eg, encoder, decoder)
        Returns:
            Dict[str, torch.Tensor]
                Dictionary mapping file names to torch tensors to be exported. If a value is
                not applicable, it will have a defined key in the produced dictionary & map to
                None.
        """
        prompt_dict = {
            PeftPromptTuning._ENCODER_KEY.name: None,
            PeftPromptTuning._DECODER_KEY.name: None,
        }
        num_transformer_submodules = model.peft_config[
            cls._ADAPTER_NAME
        ].num_transformer_submodules
        num_virtual_tokens = model.peft_config[cls._ADAPTER_NAME].num_virtual_tokens
        # Our model should only have one or two transformer modules; PEFT config lets you
        # arbitrarily configure these, but the slicing assumptions for the prompt tuning
        # seem to assume this...
        error.value_check(
            "<NLP83837722E>",
            1 <= num_transformer_submodules <= 2,
            f"Only 1 or 2 transformer submodules allowed. {num_transformer_submodules} detected.",
        )
        # Get the prompt vectors.
        if tuning_type == TuningType.PROMPT_TUNING:  # Should also be done for prefix
            # NOTE; If this is done for MPT, we get the SHARED prompt vector.
            # be careful with this, because it's the same shape as the task
            # specific tuned thing we want, and will give you garbage if you
            # leverage it directly in TGIS.
            log.info("Extracting prompt vector for prompt tuning")
            prompt_vector = model.get_prompt_embedding_to_save(
                adapter_name=cls._ADAPTER_NAME
            )
        elif tuning_type == TuningType.MULTITASK_PROMPT_TUNING:
            # For MPT / Multiprefix, run the prompt encoder, with task IDs None;
            # This assumes a single target task and produces the Hadamard product
            # of the shared prompt vector and the task learned component for Task ID 0,
            # I.e., the only task.
            prompt_tokens = (
                model.prompt_tokens[cls._ADAPTER_NAME]
                .unsqueeze(0)
                .expand(1, -1)
                .to(model.device)
            )

            log.info("Calculating single target task prompt vector")
            # Since this is running essentially an dummy forward, pass in
            # task ids as zero Tensor to forward function
            task_ids = torch.zeros(prompt_tokens.shape[0], dtype=torch.long).to(
                model.device
            )
            prompt_vector = torch.squeeze(
                model.prompt_encoder[cls._ADAPTER_NAME].forward(
                    prompt_tokens, task_ids=task_ids
                ),
                dim=0,
            )
        # Ensure that our prompt vector is on the same device as our model
        prompt_vector = prompt_vector.to(model.device)
        # Each transformer submodule should have num_virtual_tokens rows
        error.value_check(
            "<NLP83444722E>",
            prompt_vector.shape[0] == num_transformer_submodules * num_virtual_tokens,
            f"Row mismatch: Expected num_transformer_submodules * num_virtual_tokens "
            f"({num_transformer_submodules * num_virtual_tokens}) "
            f"but got {prompt_vector.shape[0]}",
        )

        # Otherwise it depends on the number of transformer modules. See seq2seq forward()

        # For Causal-LM we will essentially consider entire matrix, which has num_virtual_tokens
        # rows as the output and for Seq2Seq we will consider the 1st half of the matrix, which
        # currently has duplicate values in the second half.

        for output_type in output_model_types:
            prompt_dict[output_type.name] = prompt_vector[:num_virtual_tokens]

        return prompt_dict

    @classmethod
    def create_dataloaders_from_stream(
        cls,
        base_model: "caikit_nlp.resources.pretrained_model.base.PretrainedModelBase",
        task_type: str,
        train_stream: DataStream[GenerationTrainRecord],
        verbalizer: str,
        batch_size: int,
        max_source_length: int,
        max_target_length: int,
        validation_stream: Union[DataStream[GenerationTrainRecord], None] = None,
        collate_fn: Callable = None,
    ) -> Tuple[DataLoader]:
        """Build PyTorch data loaders around training and (optionally) evaluation DataStreams.

        Args:
            base_model: caikit_nlp.resources.pretrained_model.base.PretrainedModelBase
                Base resource model used for underlying generation.
            task_type: str
                Str indicating which task is being accomplished; currently used for determining
                tokenization / preprocessing behavior.
            train_stream: DataStream[GenerationTrainRecord]
                Data to be used for training the prompt vectors of the generation model.
            verbalizer: str
                Verbalizer template with which we will render text at both train & inference time.
            batch_size: int
                Batch size to be used for train/eval data loaders.
            max_source_length: int
                Maximum length to be used for tokenized sequences.
            max_target_length: int
                Max length of target sequences being predicted.
            validation_stream: Union[DataStream[GenerationTrainRecord], None]
                Data to be used for validation throughout the train process or None.
            collate_fn: Callable
                Function to be used for forming batches via lists of dataset inputs.

        Returns:
            Tuple[torch.utils.data.DataLoader]
                Training & evaluation datastreams for the provided data, respectively. If no
                validation_stream is provided, the returned loader for validation_stream will
                be None.
        """
        if collate_fn is None:
            # collate_fn -> pads and maps our inputs to PyTorch vectors
            collate_fn = cls._get_collate_fn(base_model.tokenizer, task_type)

        # Grab the data loaders for this task.
        # NOTE: Currently we do not expose the buffer size and we
        # default to loading the whole dataset into memory
        train_dataloader = cls._get_data_loaders_from_stream(
            base_model,
            train_stream,
            base_model.tokenizer,
            batch_size,
            collate_fn,
            verbalizer,
            max_source_length,
            max_target_length,
            shuffle=True,
        )
        if validation_stream is not None:
            val_dataloader = cls._get_data_loaders_from_stream(
                base_model,
                validation_stream,
                base_model.tokenizer,
                batch_size,
                collate_fn,
                verbalizer,
                max_source_length,
                max_target_length,
                shuffle=False,
            )
        else:
            val_dataloader = None

        return train_dataloader, val_dataloader

    @classmethod
    def create_hf_tuning_config(
        cls,
        base_model,
        tuning_type: TuningType,
        task_type: str,
        tokenizer_name_or_path: str,
        tuning_config: TuningConfig,
        output_model_types: List[PromptOutputModelType],
    ) -> PromptTuningConfig:
        """Creates Huggingface PromptTuningConfig from Caikit tuning configuration.

        Args:
            base_model: PretrainedModelBase
                Base model resource used for prompt tuning
            tuning_type: TuningType
                Type of Peft Tuning config which we would like to build.
            task_type: str
                String identifier for peft.TaskType enum, e.g., SEQ_2_SEQ_LM, CAUSAL_LM.
            tokenizer_name_or_path: str
                Name or path to the tokenizer to be leveraged.
            tuning_config: TuningConfig
                Additional model tuning configurations to be considered for prompt vector
                initialization and training behavior.
            output_model_types: List[PromptOutputModelType]
                List of output model types supported

        Returns:
            peft.PromptTuningConfig
                Peft config to be used for initializing single/multi prompt tuning.
        """

        # NOTE: Should num_virtual_tokens be part of direct `train` function param instead
        # of tuning_config?
        # NOTE: We are currently not supporting random initialization, i.e prompt_tuning_init.Random
        error.type_check("<NLP61851758E>", str, task_type=task_type)
        error.type_check("<NLP37352293E>", TuningConfig, tuning_config=tuning_config)

        error.value_check(
            "<NLP11369136E>",
            tuning_config.num_virtual_tokens
            and isinstance(tuning_config.num_virtual_tokens, int),
            "num_virtual_tokens not provided in tuning_config",
        )

        config_kwargs = tuning_config.to_dict()
        # NOTE: We are doing the mapping of state_dict_path to init_source_model
        # because we have renamed the name of that parameter in our API
        config_kwargs[
            "prompt_tuning_init_state_dict_path"
        ] = tuning_config.prompt_tuning_init_source_model

        task_type_hf = TaskType(task_type)

        config_kwargs["tokenizer_name_or_path"] = tokenizer_name_or_path
        config_kwargs[
            "num_transformer_submodules"
        ] = base_model.get_num_transformers_submodules(output_model_types)

        if tuning_config.prompt_tuning_init_method:
            config_kwargs[
                "prompt_tuning_init"
            ] = tuning_config.prompt_tuning_init_method

        if tuning_config.prompt_tuning_init_text:
            config_kwargs[
                "prompt_tuning_init_text"
            ] = tuning_config.prompt_tuning_init_text

        if tuning_type == TuningType.PROMPT_TUNING:
            tuning_config_type = PromptTuningConfig
        # elif tuning_type == TuningType.PREFIX_TUNING:
        #     tuning_config_type = PrefixTuningConfig
        elif tuning_type == TuningType.MULTITASK_PROMPT_TUNING:
            tuning_config_type = MultitaskPromptTuningConfig

        config_params = cls._filter_params_for_prompt_config(
            tuning_config_type, config_kwargs
        )
        log.info("<NLP41038481I>", f"Parameters used: {config_params}")
        return tuning_config_type(task_type=task_type_hf, **config_params)

    ################################## Private Functions ###########################################

    @classmethod
    def _get_device(cls, device: Optional[Union[str, int]]) -> Union[str, int, None]:
        """Get the device which we expect to run our models on. Defaults to GPU
        if one is available, otherwise falls back to None (cpu).

        Args:
            device: Optional[Union[str, int]]
                Device to be leveraged; if set to cls._DETECT_DEVICE, infers the device,
                otherwise we simply echo the value, which generally indicates a user override.

        Returns:
            Union[str, int, None]
                Device string that we should move our models / tensors .to() at training
                and inference time.
        """
        if device == cls._DETECT_DEVICE:
            device = "cuda" if torch.cuda.is_available() else None
            log.debug("Using device: %s", device)
        return device

    # pylint: disable=unused-argument
    @staticmethod
    def _get_collate_fn(tokenizer: AutoTokenizer, task_type: str) -> Callable:
        """Simple layer of indirection in case we want to patch in additional collate functions
        easily. Currently we always fall back to the simple default in Transformers.

        args:
            tokenizer: AutoTokenizer
                Model tokenizer. Currently this is not used, but we pass it anyway in case
                additional collate_fns dependent on it are implemented here.
            task_type: str
                Task type to be used for data collation; used for data collator overrides.

        Returns:
            Callable
                collate_fn to be used for processing batches from our datasets.
        """
        # HACK: Do NOT use the causal LM collator (for now) because
        # want to set labels ourselves. TODO: centralize collator management.
        return default_data_collator

    @staticmethod
    def _get_data_loaders_from_stream(
        base_model: PretrainedModelBase,
        train_stream: DataStream[GenerationTrainRecord],
        tokenizer: AutoTokenizer,
        batch_size: int,
        collate_fn: Callable,
        verbalizer: str,
        max_source_length: int,
        max_target_length: int,
        shuffle: bool,
    ) -> DataLoader:
        """Get the data loaders for train / evaluation.
        Args:
            base_model: caikit_nlp.resources.pretrained_model.base.PretrainedModelBase
                Base resource model used for underlying generation.
            train_stream: DataStream[GenerationTrainRecord]
                Data to be used for training the prompt vectors of the generation model.
            tokenizer: AutoTokenizer
                Model tokenizer to be used in preprocessing, i.e., when we iterate over our data.
            batch_size: int
                Batch sized to be used when building the DataLoader around the stream.
            collate_fn: Callable
                Function to be used for forming batches via lists of dataset inputs.
            verbalizer: str
                Verbalizer template to be used for formatting data. This template may use brackets
                to indicate where fields from the data model TrainGenerationRecord must be rendered.
            max_source_length: int
                Max length of sequences being considered.
            max_target_length: int
                Max length of target sequences being predicted.
            shuffle: bool
                Indicates whether or not the stream should reshuffle upon reentry.

        Returns:
            torch.utils.data.DataLoader
                DataLoader to be used for training / evaluating the stream data.
        """
        (tokenize_function, _,) = base_model.build_task_tokenize_closure(
            tokenizer, max_source_length, max_target_length, verbalizer, task_ids=0
        )
        mapped_stream = train_stream.map(tokenize_function)
        # TODO: Deprecate and remove stream wrapper & use trainer
        wrapped_stream = SimpleIterableStreamWrapper(mapped_stream, shuffle=shuffle)
        dataloader = DataLoader(
            wrapped_stream, collate_fn=collate_fn, batch_size=batch_size
        )

        return dataloader

    @classmethod
    def _execute_train_loop(
        cls,
        model: PeftModel,
        num_epochs: int,
        train_dataloader: DataLoader,
        device: str,
        eval_dataloader: Union[DataLoader, None] = None,
        metric: Optional[Callable] = None,
        learning_rate: float = 1e-3,
        tokenizer: Union[AutoTokenizer, None] = None,
        accumulate_steps: int = 1,
        silence_progress_bars: bool = True,
        torch_dtype: "torch.dtype" = torch.float32,
    ) -> None:
        """Execute the core training logic for training the prompt vectors on the frozen model.
        Note that this is done by reference.

        Args:
            model: PeftModel
                Underlying model being leveraged for text generation via prompt tuning.
            num_epochs: int
                Number of epochs to train.
            train_dataloader: torch.utils.data.DataLoader
                DataLoader to be used for loading training data.
            device: str
                Device to be used for training the model.
            eval_dataloader: Union[DataLoader, None].
                DataLoader to be used for loading eval data or None.
            metric: Union[Callable, None]
                Function to be used for evaluating data if an eval data loader is provided.
                Default: None.
            learning_rate: float
                Learning rate to be used while tuning prompt vectors. Default: 1e-3.
            tokenizer: Union[AutoTokenizer, None]
                Tokenizer for default evaluation; only used if no metric is provided and we have
                an eval dataloader.
                TODO - remove this can likely be removed.
            accumulate_steps: int
                Number of steps to use for gradient accumulation. Default: 1.
            silence_progress_bars: bool
                Silences TQDM progress bars. Default: True
            torch_dtype: torch.dtype
                Dtype to be used for training. Default: torch.float32

        Returns:
            training_metadata: Dict
                Metadata computed during training
        """
        optimizer = AdamW(params=model.parameters(), lr=learning_rate)
        lr_scheduler = get_linear_schedule_with_warmup(
            optimizer=optimizer,
            num_warmup_steps=0,
            num_training_steps=(len(train_dataloader) * num_epochs),
        )

        # Enable gradient checkpointing
        model.gradient_checkpointing_enable()

        if torch_dtype == torch.float16:
            mixed_precision = "fp16"
        elif (
            torch.cuda.is_available()
            and torch.cuda.is_bf16_supported()
            and torch_dtype == torch.bfloat16
        ):
            mixed_precision = "bf16"
        else:
            mixed_precision = "no"

        accelerator = Accelerator(
            gradient_accumulation_steps=accumulate_steps,
            device_placement=True,
            mixed_precision=mixed_precision,
        )

        # Disable cache for training
        model.config.use_cache = False

        # Below would send all the data and model to
        # configured device and convert them to required dtypes
        model, optimizer, new_train_dataloader, lr_scheduler = accelerator.prepare(
            model,
            optimizer,
            train_dataloader,
            lr_scheduler,
        )

        training_loss_tracker = []

        step_count = 1

        for epoch in range(num_epochs):
            step_loss_log = {}
            model.train()
            total_loss = 0
            tqdm_loader = tqdm(new_train_dataloader, disable=silence_progress_bars)
            for batch in tqdm_loader:

                tqdm_loader.set_description("Epoch: {}".format(epoch))

                # TODO Can this dict comprehension always replace "batch.to(device)" for us?
                try:
                    with accelerator.accumulate(model):
                        outputs = model(**batch)
                        loss = outputs.loss
                        # We are converting loss to float explicitely for later use
                        # keeping it in tensor form can potentially cause memory issues
                        loss_float = loss.detach().float().item()
                        total_loss += loss_float
                        accelerator.backward(loss)
                        optimizer.step()
                        lr_scheduler.step()
                        optimizer.zero_grad()
                        step_loss_log[step_count] = loss_float
                        step_count += 1
                except (
                    torch.cuda.OutOfMemoryError  # pylint: disable=catching-non-exception
                ):
                    error(
                        "<NLP07175292E>",
                        MemoryError("Not enough memory available for training!"),
                    )

            log.info("<NLP46114010I>", {"loss": loss_float, "epoch": epoch})

            for step, loss_val in step_loss_log.items():

                # Below is added to be propagated and stored as training_metadata
                training_loss_tracker.append(
                    {
                        "epoch": epoch,
                        "step": step,
                        "value": loss_val,
                        "timestamp": datetime.isoformat(datetime.now()),
                    }
                )

            if eval_dataloader is not None:
                model.eval()

                if metric is not None:
                    for _, batch in enumerate(
                        tqdm(eval_dataloader, disable=silence_progress_bars)
                    ):
                        batch.to(device)
                        with torch.no_grad():
                            outputs = model(**batch)
                        predictions = outputs.logits.argmax(dim=-1)
                        references = batch["labels"]
                        metric.add_batch(
                            predictions=predictions,
                            references=references,
                        )
                    eval_metric = metric.compute()

                    log.info("epoch %s: %s", epoch, eval_metric)
                else:
                    eval_loss = 0
                    # TODO Can we get away with not maintaining eval_preds?
                    eval_preds = []
                    for _, batch in enumerate(
                        tqdm(eval_dataloader, disable=silence_progress_bars)
                    ):
                        batch = {k: v.to(device) for k, v in batch.items()}
                        with torch.no_grad():
                            outputs = model(**batch)
                        loss = outputs.loss
                        eval_loss += loss.detach().float()

                        if tokenizer is not None:
                            eval_preds.extend(
                                tokenizer.batch_decode(
                                    torch.argmax(outputs.logits, -1)
                                    .detach()
                                    .cpu()
                                    .numpy(),
                                    skip_special_tokens=True,
                                )
                            )

                    eval_epoch_loss = eval_loss / len(train_dataloader)
                    eval_ppl = torch.exp(eval_epoch_loss)
                    train_epoch_loss = total_loss / len(eval_dataloader)
                    train_ppl = torch.exp(train_epoch_loss)
                    log.debug(
                        "epoch %s: %s %s %s %s",
                        epoch,
                        train_ppl,
                        train_epoch_loss,
                        eval_ppl,
                        eval_epoch_loss,
                    )
        return {"loss": training_loss_tracker}

    @classmethod
    def _filter_params_for_prompt_config(cls, prompt_config, params):
        """Utility function to filter out required parameters for prompt_config
        from `params`

        Args:
            prompt_config: PromptTuningConfig
                Tuning config type, eg:, PromptTuningConfig
            params: dict
                Dictionary containing all the input training params

        Returns:
            dict:
                Dictionary containing required params for prompt_config
        """
        # Inspect the underlying dataclass fileds; we do this because the common super class
        # used for multi/vanilla prompt/prefix tuning is a DataClass; we can't use __dict__
        # because the dataclass fields are omitted.
        allowed_keys = list(prompt_config.__dataclass_fields__.keys())
        allowed_params = dict(filter(lambda x: x[0] in allowed_keys, params.items()))
        log.info(
            "<NLP18184771I>",
            "[{}] config params not supported by provided tuning type!".format(
                params.keys() - allowed_params.keys()
            ),
        )
        return allowed_params

    @staticmethod
    def convert_peft_model_to_type(
        device: str, peft_model: PeftModel, torch_dtype=Union[str, torch.dtype]
    ) -> None:
        """Convert the underlying data type of this model to the passed dtype.

        Args:
            device: str
                Device to move our model onto.
            peft_model: PeftModel
                Model to be moved and converted to the provided PyTorch type.
            torch_dtype: Union[str, torch.dtype]
                Torch data type that we would like to coerce our model reference to.
        """
        error.type_check("<NLP83837212E>", str, allow_none=True, device=device)
        error.type_check("<NLP83837222E>", PeftModel, peft_model=peft_model)
        error.type_check("<NLP83837232E>", torch.dtype, str, torch_dtype=torch_dtype)
        # Get the actual torch type and validate, e.g., if we passed a string,
        # then move the peft model to that type on our training device.
        torch_dtype = get_torch_dtype(torch_dtype)
        # If our requested dtype is bfloat16 & we don't support it, fall back to float32
        if torch_dtype == torch.bfloat16 and (
            device == "cpu" or not torch.cuda.is_bf16_supported()
        ):
            log.warning(
                "<NLP18555772W>",
                "Requested data type torch.bfloat16 is unsupported; falling back to torch.float32",
            )
            torch_dtype = torch.float32
        peft_model.to(device, torch_dtype)



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


def env_val_to_bool(val):
    """Returns the bool value of env var"""
    if val is None:
        return False
    if isinstance(val, bool):
        return val

    # For testing env vars for values that mean false (else True!)
    return str(val).lower().strip() not in ("no", "n", "false", "0", "f", "off", "")


def env_val_to_int(val, default):
    """Returns the integer value of env var or default value if None or invalid integer"""
    try:
        return int(val)
    except (TypeError, ValueError):
        return default



"""Helpful fixtures for configuring individual unit tests."""

# Standard
from contextlib import contextmanager
from typing import Iterable, Optional, Union
from unittest import mock
import json
import os
import random
import tempfile

# Third Party
from datasets import load_dataset
import numpy as np
import pytest
import torch
import transformers

# First Party
from caikit.config.config import merge_configs
from caikit.interfaces.nlp.data_model import GeneratedTextResult, TokenizationResults
from caikit_tgis_backend import TGISBackend
from caikit_tgis_backend.tgis_connection import TGISConnection
import aconfig
import caikit

# Local
from caikit_nlp.resources.pretrained_model import HFAutoCausalLM, HFAutoSeq2SeqLM
import caikit_nlp

### Constants used in fixtures
FIXTURES_DIR = os.path.join(os.path.dirname(__file__))
TINY_MODELS_DIR = os.path.join(FIXTURES_DIR, "tiny_models")
SEQ_CLASS_MODEL = os.path.join(TINY_MODELS_DIR, "BertForSequenceClassification")
CAUSAL_LM_MODEL = os.path.join(TINY_MODELS_DIR, "BloomForCausalLM")
SEQ2SEQ_LM_MODEL = os.path.join(TINY_MODELS_DIR, "T5ForConditionalGeneration")


@pytest.fixture()
def set_cpu_device(request):
    """Fixture to set default cuda device.
    This fixture is particularly useful for running the unit tests where
    cuda devices are available, in which case, some transformers function
    may try to consume cuda and give device mismatch error.
    """
    visible_devices = os.environ.get("CUDA_VISIBLE_DEVICES", "")
    os.environ["CUDA_VISIBLE_DEVICES"] = ""
    with mock.patch.object(torch.cuda, "is_available", return_value=False):
        yield
    os.environ["CUDA_VISIBLE_DEVICES"] = visible_devices


@pytest.fixture
def disable_wip(request):
    """Fixture to temporarily disable wip decorator"""
    previous_state = caikit.core.toolkit.wip_decorator._ENABLE_DECORATOR
    caikit.core.toolkit.wip_decorator.disable_wip()
    yield
    caikit.core.toolkit.wip_decorator._ENABLE_DECORATOR = previous_state


### Fixtures for downloading objects needed to run tests via public internet
@pytest.fixture
def downloaded_dataset(request):
    """Download a dataset via datasets; this should be called via indirect parameterization.

    Example:
    @pytest.mark.parametrize(
        'download_dataset',
        [{"dataset_path": "ought/raft", "dataset_name": "twitter_complaints"}],
        indirect=True
    )
    def test_download_data(downloaded_dataset):
        # downloaded_dataset is a reference to the loaded dataset

    NOTE: Currently this is unused, but we keep it here in case we get some tiny
    models that we want to run quick regression tests on in the future.
    """
    dataset_path = request.param["dataset_path"]
    dataset_name = request.param["dataset_name"]
    return load_dataset(dataset_path, dataset_name)


@pytest.fixture
def temp_cache_dir(request):
    """Use a temporary directory as our transformers / huggingface cache dir. We have permission
    to do things in this directory, but we don't have anything downloaded in it.
    """
    old_cache_path = transformers.utils.hub.TRANSFORMERS_CACHE
    # Create a new tempdir & cache it
    temp_dir = tempfile.TemporaryDirectory()
    transformers.utils.hub.TRANSFORMERS_CACHE = temp_dir.name
    yield
    temp_dir.cleanup()
    if old_cache_path is not None:
        transformers.utils.hub.TRANSFORMERS_CACHE = old_cache_path


@pytest.fixture
def models_cache_dir(request):
    """Use the tiny models directory as the HuggingFace Cache."""
    old_cache_path = transformers.utils.hub.TRANSFORMERS_CACHE
    transformers.utils.hub.TRANSFORMERS_CACHE = TINY_MODELS_DIR
    yield
    if old_cache_path is not None:
        transformers.utils.hub.TRANSFORMERS_CACHE = old_cache_path


### Fixtures for grabbing a randomly initialized model to test interfaces against
## Causal LM
@pytest.fixture(scope="session")
def causal_lm_train_kwargs():
    """Get the kwargs for a valid train call to a Causal LM."""
    model_kwargs = {
        "base_model": HFAutoCausalLM.bootstrap(
            model_name=CAUSAL_LM_MODEL, tokenizer_name=CAUSAL_LM_MODEL
        ),
        "train_stream": caikit.core.data_model.DataStream.from_iterable(
            [
                caikit_nlp.data_model.GenerationTrainRecord(
                    input="@foo what a cute dog!", output="no complaint"
                ),
            ]
        ),
        "num_epochs": 0,
        "tuning_config": caikit_nlp.data_model.TuningConfig(
            num_virtual_tokens=8, prompt_tuning_init_text="hello world"
        ),
    }
    return model_kwargs


@pytest.fixture(scope="session")
def causal_lm_dummy_model(causal_lm_train_kwargs):
    """Train a Causal LM dummy model."""
    return caikit_nlp.modules.text_generation.PeftPromptTuning.train(
        **causal_lm_train_kwargs
    )


@pytest.fixture(scope="session")
def saved_causal_lm_dummy_model(causal_lm_dummy_model):
    """Give a path to a saved dummy model that can be loaded"""
    with tempfile.TemporaryDirectory() as workdir:
        model_dir = os.path.join(workdir, "dummy-model")
        causal_lm_dummy_model.save(model_dir)
        yield model_dir


## Seq2seq
@pytest.fixture(scope="session")
def seq2seq_lm_train_kwargs():
    """Get the kwargs for a valid train call to a Causal LM."""
    model_kwargs = {
        "base_model": HFAutoSeq2SeqLM.bootstrap(
            model_name=SEQ2SEQ_LM_MODEL, tokenizer_name=SEQ2SEQ_LM_MODEL
        ),
        "train_stream": caikit.core.data_model.DataStream.from_iterable(
            [
                caikit_nlp.data_model.GenerationTrainRecord(
                    input="@foo what a cute dog!", output="no complaint"
                ),
            ]
        ),
        "num_epochs": 0,
        "tuning_config": caikit_nlp.data_model.TuningConfig(
            num_virtual_tokens=16, prompt_tuning_init_text="hello world"
        ),
    }
    return model_kwargs


@pytest.fixture(scope="session")
def seq2seq_lm_dummy_model(seq2seq_lm_train_kwargs):
    """Train a Seq2Seq LM dummy model."""
    return caikit_nlp.modules.text_generation.PeftPromptTuning.train(
        **seq2seq_lm_train_kwargs
    )


@pytest.fixture()
def requires_determinism(request):
    """Set all of random seeds for tests expected to be deterministic."""
    # Basically, set the random seed of anything our tests might depend on
    seed = 1
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    transformers.set_seed(seed)


### Common TGIS stub classes


# Helper stubs / mocks; we use these to patch caikit so that we don't actually
# test the TGIS backend directly, and instead stub the client and inspect the
# args that we pass to it.
class StubTGISClient:
    def __init__(self, base_model_name):
        pass

    def Generate(self, request, **kwargs):
        return StubTGISClient.unary_generate(request)

    def GenerateStream(self, request, **kwargs):
        return StubTGISClient.stream_generate(request)

    def Tokenize(self, request, **kwargs):
        return StubTGISClient.tokenize(request)

    @staticmethod
    def unary_generate(request, **kwargs):
        fake_response = mock.Mock()
        fake_result = mock.Mock()
        fake_result.stop_reason = 5
        fake_result.generated_token_count = 1
        fake_result.text = "moose"
        fake_result.input_token_count = 1
        token = mock.Mock()
        token.text = "moose"
        token.logprob = 0.2
        token.rank = 1
        fake_result.tokens = [token]
        input_tokens = mock.Mock()
        input_tokens.text = "moose"
        input_tokens.logprob = 0.2
        input_tokens.rank = 1
        fake_result.input_tokens = [input_tokens]
        fake_response.responses = [fake_result]
        return fake_response

    @staticmethod
    def stream_generate(request, **kwargs):
        fake_stream = mock.Mock()
        fake_stream.stop_reason = 5
        fake_stream.generated_token_count = 1
        fake_stream.seed = 10
        fake_stream.input_token_count = 1
        token = mock.Mock()
        token.text = "moose"
        token.logprob = 0.2
        token.rank = 1
        fake_stream.tokens = [token]
        input_tokens = mock.Mock()
        input_tokens.text = "moose"
        input_tokens.logprob = 0.2
        input_tokens.rank = 1
        fake_stream.input_tokens = [input_tokens]
        fake_stream.text = "moose"
        for _ in range(3):
            yield fake_stream

    @staticmethod
    def tokenize(request, **kwargs):
        fake_response = mock.Mock()
        fake_result = mock.Mock()
        fake_result.token_count = 1
        fake_response.responses = [fake_result]
        return fake_response

    @staticmethod
    def validate_unary_generate_response(result):
        assert isinstance(result, GeneratedTextResult)
        assert result.generated_text == "moose"
        assert result.generated_tokens == 1
        assert result.finish_reason == 5
        assert result.input_token_count == 1
        assert result.tokens[0].text == "moose"
        assert result.tokens[0].logprob == 0.2
        assert result.tokens[0].rank == 1
        assert result.input_tokens[0].text == "moose"
        assert result.input_tokens[0].logprob == 0.2
        assert result.input_tokens[0].rank == 1

    @staticmethod
    def validate_stream_generate_response(stream_result):
        assert isinstance(stream_result, Iterable)
        # Convert to list to more easily check outputs
        result_list = list(stream_result)
        assert len(result_list) == 3
        first_result = result_list[0]
        assert first_result.generated_text == "moose"
        assert first_result.tokens[0].text == "moose"
        assert first_result.tokens[0].logprob == 0.2
        assert first_result.tokens[0].rank == 1
        assert first_result.input_tokens[0].text == "moose"
        assert first_result.input_tokens[0].logprob == 0.2
        assert first_result.input_tokens[0].rank == 1
        assert first_result.details.finish_reason == 5
        assert first_result.details.generated_tokens == 1
        assert first_result.details.seed == 10
        assert first_result.details.input_token_count == 1

    @staticmethod
    def validate_tokenize_response(result):
        assert isinstance(result, TokenizationResults)
        assert result.token_count == 1


class StubTGISBackend(TGISBackend):
    def __init__(
        self,
        config: Optional[dict] = None,
        temp_dir: Optional[str] = None,
        mock_remote: bool = False,
    ):
        self._temp_dir = temp_dir
        if mock_remote:
            config = config or {}
            config.update({"connection": {"hostname": "foo.{model_id}:123"}})
        super().__init__(config)
        self.load_prompt_artifacts = mock.MagicMock()
        self.unload_prompt_artifacts = mock.MagicMock()

    def get_client(self, base_model_name):
        self._model_connections[base_model_name] = TGISConnection(
            hostname="foo.bar",
            model_id=base_model_name,
            prompt_dir=self._temp_dir,
        )
        return StubTGISClient(base_model_name)


@pytest.fixture
def stub_tgis_backend():
    with tempfile.TemporaryDirectory() as temp_dir:
        yield StubTGISBackend(temp_dir=temp_dir)


### Args for commonly used datasets that we can use with our fixtures accepting params
TWITTER_DATA_DOWNLOAD_ARGS = [
    {"dataset_path": "ought/raft", "dataset_name": "twitter_complaints"}
]


@contextmanager
def temp_config(**overrides):
    local_config = aconfig.Config(
        json.loads(json.dumps(caikit.config.get_config())),
        override_env_vars=False,
    )
    merge_configs(local_config, overrides)

    with mock.patch.object(caikit.config.config, "_IMMUTABLE_CONFIG", local_config):
        yield local_config


class TestServicerContext:
    """
    A dummy class for mimicking ServicerContext invocation metadata storage.
    """

    def __init__(self, metadata: dict[str, Union[str, bytes]]):
        self.metadata = metadata

    def invocation_metadata(self):
        return list(self.metadata.items())



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# Standard
import re

# First Party
from caikit.core.exceptions import error_handler
import alog

log = alog.use_channel("VERBALIZER_UTIL")
error = error_handler.get(log)


def is_valid_verbalizer(verbalizer_template: str) -> bool:
    """Given a verbalizer template, determine if it's valid or not. We say a verbalizer
    template is valid if and only if we have at least one renderable field.

    Args:
        verbalizer_template: str
            Tentative verbalizer template to be used in text generation.
    Returns:
        bool:
            True if this is a valid verbalizer str with at least one renderable placeholder.
    """
    if not isinstance(verbalizer_template, str):
        return False
    return re.search(r"{{([_a-zA-Z0-9]+)}}", verbalizer_template) is not None


def render_verbalizer(verbalizer_template: str, source_object) -> str:
    """Given a verbalizer template and a source object, replace all templates with keys or
    attributes from the source object. Templates are expected to follow Python variable name
    allowed chars, i.e., alphanumeric with underscores; if they don't, they're skipped. The
    contents of the template will be replaced with either keys on the source object, or attribute
    values.

    Templates should be in double brackets with no whitespace, e.g., {{label}}. Consider the
    following examples.

    Examples:
        1. [Dictionary based]
            verbalizer_template = "Hello {{label}}"
            source_object = {"label": "world"}
            -> replace {{label}} with source_object["label"], producing "Hello world"
            returns: "Hello world"
        2. [Object based]
            verbalizer_template = "Source: {{input}} Target: {{output}}"
            source_object = GenerationTrainRecord(input="machine", output="learning")
            -> replace {{input}} with getattr(source_object, "source") and replace
               {{output}} with getattr(source_object, "target").
            returns: "Source: machine Target: learning"

    NOTE: This function will throw  KeyError/AttributeError if you try to grab a key or property
    that is invalid.

    Args:
        verbalizer_template: str
            Verbalizer that we want to render object values into.

    Returns:
        str
            Verbalizer string with placeholders rendered.
    """
    is_dict = isinstance(source_object, dict)

    def replace_text(match_obj):
        captured_groups = match_obj.groups()
        if len(captured_groups) != 1:
            error(
                "<NLP97444192E>",
                ValueError(
                    "Unexpectedly captured multiple groups in verbalizer rendering"
                ),
            )

        index_object = captured_groups[0]
        if is_dict:
            if index_object not in source_object:
                error(
                    "<NLP97415192E>",
                    KeyError("Requested template string is not a valid key in dict"),
                )
            return source_object[index_object]

        if not hasattr(source_object, index_object):
            error(
                "<NLP97715112E>",
                AttributeError(
                    "Requested template string is not a valid property of type"
                ),
            )
        return getattr(source_object, index_object)

    return re.sub(r"{{([_a-zA-Z0-9]+)}}", replace_text, verbalizer_template)



"""Tests for prompt tuning via PEFT.

NOTE: Currently these tests mix unit tests & regression tests.
We should be aware of the performance implications here, especially
if we start running the tests on CPUs in our CI; we'll likely want
to separate these in the future.
"""
# Standard
from typing import Iterable
from unittest import mock
import os
import tempfile

# Third Party
import pytest
import torch

# First Party
from caikit.interfaces.nlp.data_model import (
    ClassificationTrainRecord,
    GeneratedTextResult,
    GeneratedTextStreamResult,
)
import caikit

# Local
from caikit_nlp.data_model import ExponentialDecayLengthPenalty
from caikit_nlp.modules.text_generation import PeftPromptTuning
from caikit_nlp.modules.text_generation.peft_prompt_tuning import TuningType
from tests.fixtures import (
    causal_lm_dummy_model,
    causal_lm_train_kwargs,
    seq2seq_lm_dummy_model,
    seq2seq_lm_train_kwargs,
    set_cpu_device,
    temp_config,
)
import caikit_nlp

# Indexes into the peft config dictionary to get the actual prompt tuning config
DEFAULT_ADAPTER = "default"


### Tests validating block interfaces and behavior
def test_save_and_reload_with_base_model(causal_lm_dummy_model, set_cpu_device):
    """Ensure that we can save a model + its base to a tempdir and reload it."""
    with tempfile.TemporaryDirectory() as model_dir:
        causal_lm_dummy_model.save(model_dir, save_base_model=True)
        reloaded_config = caikit.core.ModuleConfig.load(os.path.abspath(model_dir))
        # We should have an exported decoder, but no encoder]
        assert os.path.isfile(os.path.join(model_dir, reloaded_config.DECODER))
        assert reloaded_config.ENCODER == ""
        reloaded_model = caikit_nlp.load(model_dir, torch_dtype="float16")
        assert reloaded_model.model.dtype is torch.float16
    assert isinstance(
        reloaded_model, caikit_nlp.modules.text_generation.PeftPromptTuning
    )


def test_save_and_reload_without_base_model(causal_lm_dummy_model):
    """Ensure that if we don't save the base model, we get the expected behavior."""
    with tempfile.TemporaryDirectory() as model_dir:
        causal_lm_dummy_model.save(model_dir, save_base_model=False)
        # For now, if we are missing the base model at load time, we throw ValueError
        with pytest.raises(ValueError):
            caikit_nlp.load(model_dir)


def test_save_log_loss_file(causal_lm_dummy_model):
    """Ensure saving a model saves the log loss file"""
    with tempfile.TemporaryDirectory() as model_dir:
        causal_lm_dummy_model.save(model_dir, save_base_model=False)
        file_path = os.path.join(
            model_dir,
            caikit_nlp.modules.text_generation.peft_prompt_tuning.TRAINING_LOSS_LOG_FILENAME,
        )

        assert os.path.isfile(file_path)


def test_run_model(causal_lm_dummy_model):
    """Ensure that we can run a model and get the right type out."""
    pred = causal_lm_dummy_model.run("This text doesn't matter")
    assert isinstance(pred, GeneratedTextResult)


def test_run_stream_out_model(causal_lm_dummy_model):
    """Ensure that we can run output streaming on a model and get the right type out."""
    pred_stream = causal_lm_dummy_model.run_stream_out("This text doesn't matter")
    assert isinstance(pred_stream, Iterable)
    for pred in pred_stream:
        print(pred)
        assert isinstance(pred, GeneratedTextStreamResult)


def test_verbalizer_rendering(causal_lm_dummy_model, monkeypatch):
    """Ensure that our model renders its verbalizer text correctly before calling tokenizer."""
    # Mock the tokenizer; we want to make sure its inputs are rendered properly
    monkeypatch.setattr(
        causal_lm_dummy_model,
        "tokenizer",
        mock.Mock(
            side_effect=RuntimeError("Tokenizer is a mock!"),
            # Set eos token property to be attribute of tokenizer
            eos_token="</s>",
        ),
    )
    input_text = "This text doesn't matter"
    causal_lm_dummy_model.verbalizer = " | {{input}} |"
    expected_tok_input = causal_lm_dummy_model.verbalizer.replace(
        "{{input}}", input_text
    )
    with pytest.raises(RuntimeError):
        causal_lm_dummy_model.run(input_text)
    # Our tokenizer mock should have run one time before our runtime error side effect
    assert len(causal_lm_dummy_model.tokenizer.call_args_list) == 1
    actual_tok_input = causal_lm_dummy_model.tokenizer.call_args_list[0].args[0]
    assert actual_tok_input == expected_tok_input


def test_verbalizer_cannot_be_static(causal_lm_train_kwargs):
    """Ensure that we throw an error if the verbalizer has no template (static model inputs)."""
    # NOTE: num_epochs 0 is intentional here; we should not need to rely on iterating over
    # our training data to determine that our verbalizer is invalid, we should do that before
    # we enter our training loop.
    patch_kwargs = {
        "verbalizer": "This has no template, so it always renders to the same thing..."
    }
    causal_lm_train_kwargs.update(patch_kwargs)
    with pytest.raises(ValueError):
        caikit_nlp.modules.text_generation.PeftPromptTuning.train(
            **causal_lm_train_kwargs
        )


def test_train_model(causal_lm_train_kwargs, set_cpu_device):
    """Ensure that we can train a model on some toy data for 1+ steps & run inference."""
    patch_kwargs = {
        "num_epochs": 1,
        "verbalizer": "Tweet text : {{input}} Label : ",
        "train_stream": caikit.core.data_model.DataStream.from_iterable(
            [
                caikit_nlp.data_model.GenerationTrainRecord(
                    input="@foo what a cute dog!", output="no complaint"
                ),
                caikit_nlp.data_model.GenerationTrainRecord(
                    input="@bar this is the worst idea ever.", output="complaint"
                ),
            ]
        ),
        "torch_dtype": torch.bfloat16,
        "device": "cpu",
    }
    causal_lm_train_kwargs.update(patch_kwargs)
    model = caikit_nlp.modules.text_generation.PeftPromptTuning.train(
        **causal_lm_train_kwargs
    )
    # Test fallback to float32 behavior if this machine doesn't support bfloat16
    assert model.model.dtype is torch.float32
    # Ensure that we can get something out of it
    pred = model.run("@bar what a cute cat!")
    assert isinstance(pred, GeneratedTextResult)


def test_gen_trained_mpt(causal_lm_train_kwargs, set_cpu_device):
    """Ensure that we are able to do generation on causal-lm model trained
    using MPT."""
    patch_kwargs = {
        "num_epochs": 1,
        "verbalizer": "Tweet text : {{input}} Label : ",
        "train_stream": caikit.core.data_model.DataStream.from_iterable(
            [
                caikit_nlp.data_model.GenerationTrainRecord(
                    input="@foo what a cute dog!", output="no complaint"
                ),
                caikit_nlp.data_model.GenerationTrainRecord(
                    input="@bar this is the worst idea ever.", output="complaint"
                ),
            ]
        ),
        "torch_dtype": torch.float32,
        "tuning_type": "MULTITASK_PROMPT_TUNING",
        "device": "cpu",
    }
    causal_lm_train_kwargs.update(patch_kwargs)
    model = caikit_nlp.modules.text_generation.PeftPromptTuning.train(
        **causal_lm_train_kwargs
    )
    # Ensure that we can get something out of it
    pred = model.run("@bar what a cute cat!")
    assert isinstance(pred, GeneratedTextResult)


def test_train_model_classification_record(causal_lm_train_kwargs, set_cpu_device):
    """Ensure that we can train a model on some toy data for 1+ steps & run inference."""
    patch_kwargs = {
        "num_epochs": 1,
        "verbalizer": "Tweet text : {{input}} Label : ",
        "train_stream": caikit.core.data_model.DataStream.from_iterable(
            [
                ClassificationTrainRecord(
                    text="@foo what a cute dog!", labels=["no complaint"]
                ),
                ClassificationTrainRecord(
                    text="@bar this is the worst idea ever.", labels=["complaint"]
                ),
            ]
        ),
        "torch_dtype": torch.bfloat16,
        "device": "cpu",
    }
    causal_lm_train_kwargs.update(patch_kwargs)
    model = caikit_nlp.modules.text_generation.PeftPromptTuning.train(
        **causal_lm_train_kwargs
    )
    # Test fallback to float32 behavior if this machine doesn't support bfloat16
    assert model.model.dtype is torch.float32
    # Ensure that we can get something out of it
    pred = model.run("@bar what a cute cat!")
    assert isinstance(pred, GeneratedTextResult)


def test_prompt_output_types(causal_lm_train_kwargs):
    # Try training a model with output_model_types set to a list of strings
    patch_kwargs = {
        "num_epochs": 1,
        "verbalizer": "Tweet text : {{input}} Label : ",
        "train_stream": caikit.core.data_model.DataStream.from_iterable(
            [
                caikit_nlp.data_model.GenerationTrainRecord(
                    input="@foo what a cute dog!", output="no complaint"
                ),
                caikit_nlp.data_model.GenerationTrainRecord(
                    input="@bar this is the worst idea ever.", output="complaint"
                ),
            ]
        ),
        "torch_dtype": torch.bfloat16,
        "device": "cpu",
        "tuning_config": caikit_nlp.data_model.TuningConfig(
            num_virtual_tokens=8,
            prompt_tuning_init_text="hello world",
            output_model_types=["DECODER"],
        ),
    }
    causal_lm_train_kwargs.update(patch_kwargs)
    model = caikit_nlp.modules.text_generation.PeftPromptTuning.train(
        **causal_lm_train_kwargs
    )
    assert model

    patch_kwargs = {
        "tuning_config": caikit_nlp.data_model.TuningConfig(
            num_virtual_tokens=8,
            prompt_tuning_init_text="hello world",
            output_model_types=[caikit_nlp.data_model.PromptOutputModelType.DECODER],
        )
    }
    model = caikit_nlp.modules.text_generation.PeftPromptTuning.train(
        **causal_lm_train_kwargs
    )
    assert model


def test_error_empty_stream(causal_lm_train_kwargs):
    patch_kwargs = {
        "num_epochs": 1,
        "verbalizer": "Tweet text : {{input}} Label : ",
        "train_stream": caikit.core.data_model.DataStream.from_iterable([]),
    }
    causal_lm_train_kwargs.update(patch_kwargs)
    with pytest.raises(ValueError):
        caikit_nlp.modules.text_generation.PeftPromptTuning.train(
            **causal_lm_train_kwargs
        )


### Implementation details
# These tests can probably be removed and tested directly through .save() once
# full seq2seq support is completed and verified.
def test_get_prompt_vector_encoder_only_seq2seq(seq2seq_lm_dummy_model):
    """Ensure that if we have an encoder/decoder PEFT model, we can get the encoder vectors."""
    peft_model = seq2seq_lm_dummy_model.model
    # Ensure we have 16 rows in our embed matrix; otherwise the below operations are invalid
    peft_model.peft_config[DEFAULT_ADAPTER].num_virtual_tokens = 16
    peft_model.peft_config[DEFAULT_ADAPTER].num_transformer_submodules = 1

    prompt_vectors = seq2seq_lm_dummy_model.get_exportable_prompt_vectors(
        peft_model,
        seq2seq_lm_dummy_model.tuning_type,
        seq2seq_lm_dummy_model.output_model_types,
    )
    # Get the dimensions of our word embeddings
    dummy_emb_input = torch.tensor([1]).to(peft_model.device)
    word_embedding_dim = peft_model.word_embeddings(dummy_emb_input).shape[1]
    # Ensure that only the encoder is of dim (num_virtual_tokens, word_embedding_dim)
    assert prompt_vectors[PeftPromptTuning._ENCODER_KEY.name].shape == (
        peft_model.peft_config[DEFAULT_ADAPTER].num_virtual_tokens,
        word_embedding_dim,
    )
    assert prompt_vectors[PeftPromptTuning._DECODER_KEY.name] is None


@pytest.mark.skip(
    """
We are skipping this test because currently we do not have
 support for encoder-decoder output for seq2seq
"""
)
def test_get_prompt_vector_encoder_decoder_seq2seq(seq2seq_lm_dummy_model):
    """Ensure that if we have an encoder/decoder PEFT model, we can get both prompt vectors."""
    peft_model = seq2seq_lm_dummy_model.model
    # Ensure we have 16 rows in our embed matrix; otherwise the below operations are invalid
    assert (
        peft_model.peft_config[DEFAULT_ADAPTER].num_virtual_tokens
        * peft_model.peft_config[DEFAULT_ADAPTER].num_transformer_submodules
        == 16
    )
    peft_model.peft_config[DEFAULT_ADAPTER].num_virtual_tokens = 8
    peft_model.peft_config[DEFAULT_ADAPTER].num_transformer_submodules = 2

    # Ensure that if we grab the prompt from the model, we have both an encoder and decoder.
    # So our embedding matrix should be (2 * num_virtual_tokens, word_embedding_dim)
    prompt_vectors = seq2seq_lm_dummy_model.get_exportable_prompt_vectors(
        peft_model,
        seq2seq_lm_dummy_model.tuning_type,
        seq2seq_lm_dummy_model.output_model_types,
    )
    # Get the dimensions of our word embeddings
    dummy_emb_input = torch.tensor([1]).to(peft_model.device)
    word_embedding_dim = peft_model.word_embeddings(dummy_emb_input).shape[1]

    # Ensure that the encoder / decoder are each of dim (num_virtual_tokens, word_embedding_dim)
    assert prompt_vectors[PeftPromptTuning._ENCODER_KEY.name].shape == (
        peft_model.peft_config[DEFAULT_ADAPTER].num_virtual_tokens,
        word_embedding_dim,
    )
    assert prompt_vectors[PeftPromptTuning._DECODER_KEY.name].shape == (
        peft_model.peft_config[DEFAULT_ADAPTER].num_virtual_tokens,
        word_embedding_dim,
    )


def test_model_can_only_have_one_or_two_transformer_modules(seq2seq_lm_dummy_model):
    """Ensure that we explode if we have more transformer modules than we should."""
    peft_model = seq2seq_lm_dummy_model.model
    # Ensure we have 16 rows in our embed matrix; otherwise the below operations are invalid
    assert (
        peft_model.peft_config[DEFAULT_ADAPTER].num_virtual_tokens
        * peft_model.peft_config[DEFAULT_ADAPTER].num_transformer_submodules
        == 16
    )
    # Peft doesn't enforce the range 1 <= x <= 2 at initialization time, so it's good to ensure the
    # export knows how to handle it directly.
    peft_model.peft_config[DEFAULT_ADAPTER].num_virtual_tokens = 4
    peft_model.peft_config[DEFAULT_ADAPTER].num_transformer_submodules = 4
    with pytest.raises(ValueError):
        PeftPromptTuning.get_exportable_prompt_vectors(
            peft_model,
            TuningType.PROMPT_TUNING,
            seq2seq_lm_dummy_model.output_model_types,
        )


######################## Test run with optional params #####################


def test_run_repetition_penalty_0_works(causal_lm_dummy_model):
    """Ensure repetition_penalty works with 0.0 as input"""
    pred = causal_lm_dummy_model.run("This text doesn't matter", repetition_penalty=0.0)
    assert isinstance(pred, GeneratedTextResult)


def test_run_truncate_tokens_0(causal_lm_dummy_model):
    """Ensure run function accepts 0 for truncation value
    and successfully turns off truncation"""
    pred = causal_lm_dummy_model.run(
        "This text doesn't matter", truncate_input_tokens=0
    )
    assert isinstance(pred, GeneratedTextResult)


def test_run_with_preserve_input_text(causal_lm_dummy_model):
    """Ensure preserve input text removes input
    from generated output when set to False"""
    input_text = "This text doesn't matter"
    pred = causal_lm_dummy_model.run(input_text, preserve_input_text=True)
    assert input_text in pred.generated_text
    pred = causal_lm_dummy_model.run(input_text, preserve_input_text=False)
    assert input_text not in pred.generated_text


def test_run_sampling_param_ignored_greedy_decoding(causal_lm_dummy_model):
    """Ensure sampling parameter gets ignored when decoding method
    is set to GREEDY
    """
    pred = causal_lm_dummy_model.run(
        "This text doesn't matter",
        decoding_method="GREEDY",
        top_k=2,
        top_p=0.23,
        typical_p=0.23,
        temperature=0.77,
    )
    assert isinstance(pred, GeneratedTextResult)


def test_run_with_custom_stop_criteria(causal_lm_dummy_model):
    """Ensure custom stop sequences works with run"""
    pred = causal_lm_dummy_model.run(
        "This text doesn't matter",
        decoding_method="GREEDY",
        stop_sequences=["Foo", "bar"],
    )
    assert isinstance(pred, GeneratedTextResult)


def test_run_exponential_decay_len_penatly_object(causal_lm_dummy_model):
    """Ensure exponential decay len penalty works with the data model
    object
    """
    penalty = ExponentialDecayLengthPenalty(1, 0.2)
    pred = causal_lm_dummy_model.run(
        "This text doesn't matter",
        decoding_method="GREEDY",
        stop_sequences=["Foo", "bar"],
        exponential_decay_length_penalty=penalty,
    )
    assert isinstance(pred, GeneratedTextResult)


def test_run_tokenizer_not_implemented(causal_lm_dummy_model):
    with pytest.raises(NotImplementedError):
        causal_lm_dummy_model.run_tokenizer("This text doesn't matter")


######################## Test train ###############################################


def test_train_with_data_validation_raises(causal_lm_train_kwargs, set_cpu_device):
    """Check if we are able to throw error for when number of examples are more than configured limit"""
    patch_kwargs = {
        "num_epochs": 1,
        "verbalizer": "Tweet text : {{input}} Label : ",
        "train_stream": caikit.core.data_model.DataStream.from_iterable(
            [
                ClassificationTrainRecord(
                    text="@foo what a cute dog!", labels=["no complaint"]
                ),
                ClassificationTrainRecord(
                    text="@bar this is the worst idea ever.", labels=["complaint"]
                ),
            ]
        ),
        "torch_dtype": torch.bfloat16,
        "device": "cpu",
    }
    causal_lm_train_kwargs.update(patch_kwargs)

    model_name = causal_lm_train_kwargs["base_model"]._model_name
    module = caikit_nlp.modules.text_generation.PeftPromptTuning
    with temp_config(training_data_limit={module.MODULE_ID: {model_name: 1}}):
        with pytest.raises(ValueError):
            module.train(**causal_lm_train_kwargs)


def test_train_with_data_validation_success(causal_lm_train_kwargs, set_cpu_device):
    """Check if we are able to train successfully if training data is within limits"""
    patch_kwargs = {
        "num_epochs": 1,
        "verbalizer": "Tweet text : {{input}} Label : ",
        "train_stream": caikit.core.data_model.DataStream.from_iterable(
            [
                ClassificationTrainRecord(
                    text="@foo what a cute dog!", labels=["no complaint"]
                ),
                ClassificationTrainRecord(
                    text="@bar this is the worst idea ever.", labels=["complaint"]
                ),
            ]
        ),
        "torch_dtype": torch.bfloat16,
        "device": "cpu",
    }
    causal_lm_train_kwargs.update(patch_kwargs)

    model_name = causal_lm_train_kwargs["base_model"]._model_name
    module = caikit_nlp.modules.text_generation.PeftPromptTuning
    with temp_config(training_data_limit={module.MODULE_ID: {model_name: 2}}):

        model = module.train(**causal_lm_train_kwargs)
        assert model


def test_train_with_non_existent_limit_success(causal_lm_train_kwargs, set_cpu_device):
    """Check if we are able to train successfully if training data limit doesn't exist for particular model"""
    patch_kwargs = {
        "num_epochs": 1,
        "verbalizer": "Tweet text : {{input}} Label : ",
        "train_stream": caikit.core.data_model.DataStream.from_iterable(
            [
                ClassificationTrainRecord(
                    text="@foo what a cute dog!", labels=["no complaint"]
                )
            ]
        ),
        "torch_dtype": torch.bfloat16,
        "device": "cpu",
    }
    causal_lm_train_kwargs.update(patch_kwargs)

    model_name = causal_lm_train_kwargs["base_model"]._model_name
    module = caikit_nlp.modules.text_generation.PeftPromptTuning
    with temp_config(training_data_limit={module.MODULE_ID: {"foo": 2}}):

        model = module.train(**causal_lm_train_kwargs)
        assert model


def test_train_with_no_limit_for_module(causal_lm_train_kwargs, set_cpu_device):
    """Check if we are able to train successfully if training data limit doesn't exist prompt tuning module"""
    patch_kwargs = {
        "num_epochs": 1,
        "verbalizer": "Tweet text : {{input}} Label : ",
        "train_stream": caikit.core.data_model.DataStream.from_iterable(
            [
                ClassificationTrainRecord(
                    text="@foo what a cute dog!", labels=["no complaint"]
                )
            ]
        ),
        "torch_dtype": torch.bfloat16,
        "device": "cpu",
    }
    causal_lm_train_kwargs.update(patch_kwargs)

    model_name = causal_lm_train_kwargs["base_model"]._model_name
    module = caikit_nlp.modules.text_generation.PeftPromptTuning
    with temp_config(training_data_limit={}):

        model = module.train(**causal_lm_train_kwargs)
        assert model


def test_train_module_level_data_validation_raises(
    causal_lm_train_kwargs, set_cpu_device
):
    """Check if train raises with module level default configuration
    if training data is within limits and model config is not provided
    """
    patch_kwargs = {
        "num_epochs": 1,
        "verbalizer": "Tweet text : {{input}} Label : ",
        "train_stream": caikit.core.data_model.DataStream.from_iterable(
            [
                ClassificationTrainRecord(
                    text="@foo what a cute dog!", labels=["no complaint"]
                ),
                ClassificationTrainRecord(
                    text="@bar this is the worst idea ever.", labels=["complaint"]
                ),
            ]
        ),
        "torch_dtype": torch.bfloat16,
        "device": "cpu",
    }
    causal_lm_train_kwargs.update(patch_kwargs)

    module = caikit_nlp.modules.text_generation.PeftPromptTuning
    with temp_config(
        training_data_limit={module.MODULE_ID: {"__default__": 1, "foo": 2}}
    ):
        with pytest.raises(ValueError):
            module.train(**causal_lm_train_kwargs)


def test_train_module_level_data_validation_success(
    causal_lm_train_kwargs, set_cpu_device
):
    """Check if we are able to train successfully with module level default configuration
    if training data is within limits and model config present
    """
    patch_kwargs = {
        "num_epochs": 1,
        "verbalizer": "Tweet text : {{input}} Label : ",
        "train_stream": caikit.core.data_model.DataStream.from_iterable(
            [
                ClassificationTrainRecord(
                    text="@foo what a cute dog!", labels=["no complaint"]
                ),
                ClassificationTrainRecord(
                    text="@bar this is the worst idea ever.", labels=["complaint"]
                ),
            ]
        ),
        "torch_dtype": torch.bfloat16,
        "device": "cpu",
    }
    causal_lm_train_kwargs.update(patch_kwargs)

    model_name = causal_lm_train_kwargs["base_model"]._model_name
    module = caikit_nlp.modules.text_generation.PeftPromptTuning
    with temp_config(
        training_data_limit={module.MODULE_ID: {"__default__": 1, model_name: 2}}
    ):

        model = module.train(**causal_lm_train_kwargs)
        assert model


def test_train_global_default_data_validation_raises(
    causal_lm_train_kwargs, set_cpu_device
):
    """Check if train raises with global default configuration
    if training data is within limits and model config is not provided
    """
    patch_kwargs = {
        "num_epochs": 1,
        "verbalizer": "Tweet text : {{input}} Label : ",
        "train_stream": caikit.core.data_model.DataStream.from_iterable(
            [
                ClassificationTrainRecord(
                    text="@foo what a cute dog!", labels=["no complaint"]
                ),
                ClassificationTrainRecord(
                    text="@bar this is the worst idea ever.", labels=["complaint"]
                ),
            ]
        ),
        "torch_dtype": torch.bfloat16,
        "device": "cpu",
    }
    causal_lm_train_kwargs.update(patch_kwargs)

    module = caikit_nlp.modules.text_generation.PeftPromptTuning
    with temp_config(
        training_data_limit={"__default__": 1, module.MODULE_ID: {"foo": 2}}
    ):
        with pytest.raises(ValueError):
            module.train(**causal_lm_train_kwargs)


def test_train_global_default_data_validation_success(
    causal_lm_train_kwargs, set_cpu_device
):
    """Check if we are able to train successfully with global default configuration
    if training data is within limits and model config is present
    """
    patch_kwargs = {
        "num_epochs": 1,
        "verbalizer": "Tweet text : {{input}} Label : ",
        "train_stream": caikit.core.data_model.DataStream.from_iterable(
            [
                ClassificationTrainRecord(
                    text="@foo what a cute dog!", labels=["no complaint"]
                ),
                ClassificationTrainRecord(
                    text="@bar this is the worst idea ever.", labels=["complaint"]
                ),
            ]
        ),
        "torch_dtype": torch.bfloat16,
        "device": "cpu",
    }
    causal_lm_train_kwargs.update(patch_kwargs)

    model_name = causal_lm_train_kwargs["base_model"]._model_name
    module = caikit_nlp.modules.text_generation.PeftPromptTuning
    with temp_config(
        training_data_limit={"__default__": 1, module.MODULE_ID: {model_name: 2}}
    ):

        model = module.train(**causal_lm_train_kwargs)
        assert model



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""This file is for helper functions related to TGIS."""

# Standard
from typing import Iterable

# Third Party
import grpc

# First Party
from caikit import get_config
from caikit.core.exceptions import error_handler
from caikit.core.exceptions.caikit_core_exception import (
    CaikitCoreException,
    CaikitCoreStatusCode,
)
from caikit.interfaces.nlp.data_model import (
    GeneratedTextResult,
    GeneratedTextStreamResult,
    GeneratedToken,
    TokenizationResults,
    TokenStreamDetails,
)
from caikit_tgis_backend import TGISBackend
from caikit_tgis_backend.protobufs import generation_pb2
import alog

# Local
from ...data_model import ExponentialDecayLengthPenalty
from .model_run_utils import GENERATE_FUNCTION_ARGS, VALID_DECODING_METHODS

log = alog.use_channel("TGIS_UTILS")
error = error_handler.get(log)

GENERATE_FUNCTION_TGIS_ARGS = """
    {}
    preserve_input_text: bool
        Whether or not the source string should be contained in the generated output,
        e.g., as a prefix.
    input_tokens: bool
        Whether or not to include list of input tokens.
    generated_tokens: bool
        Whether or not to include list of individual generated tokens.
    token_logprobs: bool
        Whether or not to include logprob for each returned token.
        Applicable only if generated_tokens == true and/or input_tokens == true
    token_ranks: bool
        Whether or not to include rank of each returned token.
        Applicable only if generated_tokens == true and/or input_tokens == true
    include_stop_sequence: bool
        Whether or not to include stop sequence.
        If not specified, default behavior depends on server setting.
""".format(
    GENERATE_FUNCTION_ARGS
)


# Mapping from grpc status codes to caikit status codes. There is not a 1:1
# mapping at the moment, so this conversion is lossy!
GRPC_TO_CAIKIT_CORE_STATUS = {
    grpc.StatusCode.CANCELLED: CaikitCoreStatusCode.CONNECTION_ERROR,
    grpc.StatusCode.UNKNOWN: CaikitCoreStatusCode.UNKNOWN,
    grpc.StatusCode.INVALID_ARGUMENT: CaikitCoreStatusCode.INVALID_ARGUMENT,
    grpc.StatusCode.DEADLINE_EXCEEDED: CaikitCoreStatusCode.CONNECTION_ERROR,
    grpc.StatusCode.NOT_FOUND: CaikitCoreStatusCode.NOT_FOUND,
    grpc.StatusCode.ALREADY_EXISTS: CaikitCoreStatusCode.INVALID_ARGUMENT,
    grpc.StatusCode.PERMISSION_DENIED: CaikitCoreStatusCode.FORBIDDEN,
    grpc.StatusCode.RESOURCE_EXHAUSTED: CaikitCoreStatusCode.INVALID_ARGUMENT,
    grpc.StatusCode.FAILED_PRECONDITION: CaikitCoreStatusCode.INVALID_ARGUMENT,
    grpc.StatusCode.ABORTED: CaikitCoreStatusCode.CONNECTION_ERROR,
    grpc.StatusCode.OUT_OF_RANGE: CaikitCoreStatusCode.INVALID_ARGUMENT,
    grpc.StatusCode.UNIMPLEMENTED: CaikitCoreStatusCode.UNKNOWN,
    grpc.StatusCode.INTERNAL: CaikitCoreStatusCode.FATAL,
    grpc.StatusCode.UNAVAILABLE: CaikitCoreStatusCode.CONNECTION_ERROR,
    grpc.StatusCode.DATA_LOSS: CaikitCoreStatusCode.CONNECTION_ERROR,
    grpc.StatusCode.UNAUTHENTICATED: CaikitCoreStatusCode.UNAUTHORIZED,
}

# HTTP Header / gRPC Metadata key used to identify a route override
# (forwarded for API compatibility)
ROUTE_INFO_HEADER_KEY = TGISBackend.ROUTE_INFO_HEADER_KEY
get_route_info = TGISBackend.get_route_info


def raise_caikit_core_exception(rpc_error: grpc.RpcError):
    """Helper to wrap logic of converting from grpc.RpcError ->
    CaikitCoreException
    """
    caikit_status_code = GRPC_TO_CAIKIT_CORE_STATUS.get(
        rpc_error.code(), CaikitCoreStatusCode.UNKNOWN
    )
    error_message = rpc_error.details() or f"Unknown RpcError: {rpc_error}"
    raise CaikitCoreException(caikit_status_code, error_message) from rpc_error


def validate_inf_params(
    text,
    preserve_input_text,
    input_tokens,
    generated_tokens,
    token_logprobs,
    token_ranks,
    include_stop_sequence,
    eos_token,
    max_new_tokens,
    min_new_tokens,
    truncate_input_tokens,
    decoding_method,
    top_k,
    top_p,
    typical_p,
    temperature,
    seed,
    repetition_penalty,
    max_time,
    exponential_decay_length_penalty,
    stop_sequences,
):
    """Validate inference parameters

    Args:
        eos_token: str
           A special token representing the end of a sentence.
        {}
    """.format(
        GENERATE_FUNCTION_TGIS_ARGS
    )
    error.type_check("<NLP65883535E>", str, text=text)
    error.type_check("<NLP65883537E>", bool, preserve_input_text=preserve_input_text)
    error.type_check("<NLP65883538E>", bool, input_tokens=input_tokens)
    error.type_check("<NLP65883539E>", bool, generated_tokens=generated_tokens)
    error.type_check("<NLP65883540E>", bool, token_logprobs=token_logprobs)
    error.type_check("<NLP65883541E>", bool, token_ranks=token_ranks)
    error.type_check(
        "<NLP65883542E>", bool, include_stop_sequence=include_stop_sequence
    )
    error.type_check("<NLP85452188E>", str, allow_none=True, eos_token=eos_token)
    error.type_check(
        "<NLP03860681E>",
        int,
        allow_none=True,
        max_new_tokens=max_new_tokens,
        min_new_tokens=min_new_tokens,
        truncate_input_tokens=truncate_input_tokens,
        top_k=top_k,
        seed=seed,
    )

    error.value_check(
        "<NLP03521352E>",
        max_new_tokens >= min_new_tokens,
        f"Maximum new tokens [{max_new_tokens}] has to be greater than minimum new tokens \
        [{min_new_tokens}]",
    )

    error.value_check(
        "<NLP03521363E>",
        decoding_method in VALID_DECODING_METHODS,
        f"Decoding method [{decoding_method}] not in valid decoding methods: "
        f"[{VALID_DECODING_METHODS}]",
    )
    error.type_check(
        "<NLP55267524E>",
        float,
        allow_none=True,
        top_p=top_p,
        typical_p=typical_p,
        temperature=temperature,
        repetition_penalty=repetition_penalty,
        max_time=max_time,
    )
    error.type_check(
        "<NLP28185345E>",
        ExponentialDecayLengthPenalty,
        tuple,
        allow_none=True,
        exponential_decay_length_penalty=exponential_decay_length_penalty,
    )

    error.type_check_all(
        "<NLP41311584E>", str, allow_none=True, stop_sequences=stop_sequences
    )

    error.value_check(
        "<NLP28185344E>",
        not temperature or temperature >= 0.05,
        "temperature must be >= 0.05",
    )

    error.value_check(
        "<NLP28185346E>",
        not top_p or 0 < top_p <= 1.0,
        "top_p must be > 0.0 and <= 1.0",
    )

    error.value_check(
        "<NLP28185347E>", not top_k or top_k >= 0, "top_k must be strictly positive"
    )

    error.value_check(
        "<NLP28185348E>", not typical_p or typical_p <= 1.0, "typical_p must be <= 1.0"
    )

    error.value_check(
        "<NLP28185349E>",
        not repetition_penalty or repetition_penalty > 0.0,
        "repetition_penalty must be > 0.0",
    )

    if exponential_decay_length_penalty:
        if isinstance(exponential_decay_length_penalty, ExponentialDecayLengthPenalty):
            exponential_decay_length_penalty = (
                exponential_decay_length_penalty.start_index,
                exponential_decay_length_penalty.decay_factor,
            )
        error.value_check(
            "<NLP28185350E>",
            exponential_decay_length_penalty[1] >= 1.0
            and exponential_decay_length_penalty[1] <= 10.0,
            "decay_factor in exponential_decay_length_penalty must be >= 1.0 and <= 10.0",
        )

    if decoding_method == "GREEDY" and (
        temperature not in (1, None)
        or top_k not in (0, None)
        or top_p not in (1, None)
        or seed
    ):
        raise ValueError(
            "sampling parameters (temperature/top_k/top_p/typical_p/seed) aren't "
            "applicable in greedy decoding mode"
        )


def get_params(
    preserve_input_text,
    input_tokens,
    generated_tokens,
    token_logprobs,
    token_ranks,
    include_stop_sequence,
    max_new_tokens,
    min_new_tokens,
    truncate_input_tokens,
    decoding_method,
    top_k,
    top_p,
    typical_p,
    temperature,
    seed,
    repetition_penalty,
    max_time,
    exponential_decay_length_penalty,
    stop_sequences,
):
    """Get generation parameters

    Args:
        {}
    """.format(
        GENERATE_FUNCTION_TGIS_ARGS
    )

    if decoding_method == "GREEDY":
        decoding = generation_pb2.DecodingMethod.GREEDY
    elif decoding_method == "SAMPLING":
        decoding = generation_pb2.DecodingMethod.SAMPLE

    sampling_parameters = generation_pb2.SamplingParameters(
        temperature=temperature,
        top_k=top_k,
        top_p=top_p,
        typical_p=typical_p,
        seed=seed,
    )

    res_options = generation_pb2.ResponseOptions(
        input_text=preserve_input_text,
        generated_tokens=generated_tokens,
        input_tokens=input_tokens,
        token_logprobs=token_logprobs,
        token_ranks=token_ranks,
    )
    stopping = generation_pb2.StoppingCriteria(
        stop_sequences=stop_sequences,
        max_new_tokens=max_new_tokens,
        min_new_tokens=min_new_tokens,
        time_limit_millis=int(max_time * 1000) if max_time else None,
        include_stop_sequence=include_stop_sequence,
    )

    if exponential_decay_length_penalty:
        if isinstance(exponential_decay_length_penalty, ExponentialDecayLengthPenalty):
            exponential_decay_length_penalty = (
                exponential_decay_length_penalty.start_index,
                exponential_decay_length_penalty.decay_factor,
            )
        exponential_decay_length_penalty = (
            generation_pb2.DecodingParameters.LengthPenalty(
                start_index=exponential_decay_length_penalty[0],
                decay_factor=exponential_decay_length_penalty[1],
            )
        )

    decoding_parameters = generation_pb2.DecodingParameters(
        repetition_penalty=repetition_penalty,
        length_penalty=exponential_decay_length_penalty,
    )

    params = generation_pb2.Parameters(
        method=decoding,
        sampling=sampling_parameters,
        response=res_options,
        stopping=stopping,
        decoding=decoding_parameters,
        truncate_input_tokens=truncate_input_tokens,
    )
    return params


class TGISGenerationClient:
    """Client for TGIS generation calls"""

    def __init__(
        self, base_model_name, eos_token, tgis_client, producer_id, prefix_id=None
    ):
        self.base_model_name = base_model_name
        self.eos_token = eos_token
        self.tgis_client = tgis_client
        self.producer_id = producer_id
        self.prefix_id = prefix_id

        self.tgis_req_timeout = get_config().tgis_request_timeout

        if (
            not self.tgis_req_timeout
            or not isinstance(self.tgis_req_timeout, int)
            or self.tgis_req_timeout <= 0
        ):
            log.debug("<RUN57106697I>", "TGIS timeout not set")
            self.tgis_req_timeout = None

        else:
            log.debug(
                "<RUN57106696T>",
                "Setting TGIS timeout value to  %d",
                self.tgis_req_timeout,
            )

    def unary_generate(
        self,
        text,
        preserve_input_text,
        input_tokens,
        generated_tokens,
        token_logprobs,
        token_ranks,
        include_stop_sequence,
        max_new_tokens,
        min_new_tokens,
        truncate_input_tokens,
        decoding_method,
        top_k,
        top_p,
        typical_p,
        temperature,
        seed,
        repetition_penalty,
        max_time,
        exponential_decay_length_penalty,
        stop_sequences,
    ) -> GeneratedTextResult:
        """Generate unary output from model in TGIS

        Args:
            {}
        Returns:
            GeneratedTextResult
                Generated text result produced by TGIS.
        """.format(
            GENERATE_FUNCTION_TGIS_ARGS
        )

        # In case internal client is not configured - generation
        # cannot be done (individual modules may already check
        # for this)
        error.value_check(
            "<NLP72700256E>",
            self.tgis_client is not None,
            "Backend must be configured and loaded for generate",
        )

        validate_inf_params(
            text=text,
            preserve_input_text=preserve_input_text,
            input_tokens=input_tokens,
            generated_tokens=generated_tokens,
            token_logprobs=token_logprobs,
            token_ranks=token_ranks,
            include_stop_sequence=include_stop_sequence,
            eos_token=self.eos_token,
            max_new_tokens=max_new_tokens,
            min_new_tokens=min_new_tokens,
            truncate_input_tokens=truncate_input_tokens,
            decoding_method=decoding_method,
            top_k=top_k,
            top_p=top_p,
            typical_p=typical_p,
            temperature=temperature,
            seed=seed,
            repetition_penalty=repetition_penalty,
            max_time=max_time,
            exponential_decay_length_penalty=exponential_decay_length_penalty,
            stop_sequences=stop_sequences,
        )

        log.debug("Building protobuf request to send to TGIS")

        params = get_params(
            preserve_input_text=preserve_input_text,
            input_tokens=input_tokens,
            generated_tokens=generated_tokens,
            token_logprobs=token_logprobs,
            token_ranks=token_ranks,
            include_stop_sequence=include_stop_sequence,
            max_new_tokens=max_new_tokens,
            min_new_tokens=min_new_tokens,
            truncate_input_tokens=truncate_input_tokens,
            decoding_method=decoding_method,
            top_k=top_k,
            top_p=top_p,
            typical_p=typical_p,
            temperature=temperature,
            seed=seed,
            repetition_penalty=repetition_penalty,
            max_time=max_time,
            exponential_decay_length_penalty=exponential_decay_length_penalty,
            stop_sequences=stop_sequences,
        )

        gen_reqs = [generation_pb2.GenerationRequest(text=text)]
        if not self.prefix_id:
            request = generation_pb2.BatchedGenerationRequest(
                requests=gen_reqs,
                model_id=self.base_model_name,
                params=params,
            )
        else:
            request = generation_pb2.BatchedGenerationRequest(
                requests=gen_reqs,
                model_id=self.base_model_name,
                prefix_id=self.prefix_id,
                params=params,
            )

        # Currently, we send a batch request of len(x)==1, so we expect one response back
        with alog.ContextTimer(log.trace, "TGIS request duration: "):
            try:
                batch_response = self.tgis_client.Generate(
                    request, timeout=self.tgis_req_timeout
                )
            except grpc.RpcError as err:
                raise_caikit_core_exception(err)

        error.value_check(
            "<NLP38899018E>",
            len(batch_response.responses) == 1,
            f"Got {len(batch_response.responses)} responses for a single request",
        )
        response = batch_response.responses[0]

        token_list = []
        if response.tokens is not None:
            for token in response.tokens:
                token_list.append(
                    GeneratedToken(
                        text=token.text, logprob=token.logprob, rank=token.rank
                    )
                )

        input_token_list = []
        if response.input_tokens is not None:
            for token in response.input_tokens:
                input_token_list.append(
                    GeneratedToken(
                        text=token.text, logprob=token.logprob, rank=token.rank
                    )
                )

        return GeneratedTextResult(
            generated_text=response.text,
            generated_tokens=response.generated_token_count,
            finish_reason=response.stop_reason,
            producer_id=self.producer_id,
            input_token_count=response.input_token_count,
            seed=seed,
            tokens=token_list,
            input_tokens=input_token_list,
        )

    def stream_generate(
        self,
        text,
        preserve_input_text,
        input_tokens,
        generated_tokens,
        token_logprobs,
        token_ranks,
        include_stop_sequence,
        max_new_tokens,
        min_new_tokens,
        truncate_input_tokens,
        decoding_method,
        top_k,
        top_p,
        typical_p,
        temperature,
        seed,
        repetition_penalty,
        max_time,
        exponential_decay_length_penalty,
        stop_sequences,
    ) -> Iterable[GeneratedTextStreamResult]:
        """Generate stream output from model in TGIS

        Args:
            {}
        Returns:
            Iterable[GeneratedTextStreamResult]
        """.format(
            GENERATE_FUNCTION_TGIS_ARGS
        )

        # In case internal client is not configured - generation
        # cannot be done (individual modules may already check
        # for this)
        error.value_check(
            "<NLP77278635E>",
            self.tgis_client is not None,
            "Backend must be configured and loaded for generate",
        )
        log.debug("Building protobuf request to send to TGIS")

        validate_inf_params(
            text=text,
            preserve_input_text=preserve_input_text,
            input_tokens=input_tokens,
            generated_tokens=generated_tokens,
            token_logprobs=token_logprobs,
            token_ranks=token_ranks,
            include_stop_sequence=include_stop_sequence,
            eos_token=self.eos_token,
            max_new_tokens=max_new_tokens,
            min_new_tokens=min_new_tokens,
            truncate_input_tokens=truncate_input_tokens,
            decoding_method=decoding_method,
            top_k=top_k,
            top_p=top_p,
            typical_p=typical_p,
            temperature=temperature,
            seed=seed,
            repetition_penalty=repetition_penalty,
            max_time=max_time,
            exponential_decay_length_penalty=exponential_decay_length_penalty,
            stop_sequences=stop_sequences,
        )

        params = get_params(
            preserve_input_text=preserve_input_text,
            input_tokens=input_tokens,
            generated_tokens=generated_tokens,
            token_logprobs=token_logprobs,
            token_ranks=token_ranks,
            include_stop_sequence=include_stop_sequence,
            max_new_tokens=max_new_tokens,
            min_new_tokens=min_new_tokens,
            truncate_input_tokens=truncate_input_tokens,
            decoding_method=decoding_method,
            top_k=top_k,
            top_p=top_p,
            typical_p=typical_p,
            temperature=temperature,
            seed=seed,
            repetition_penalty=repetition_penalty,
            max_time=max_time,
            exponential_decay_length_penalty=exponential_decay_length_penalty,
            stop_sequences=stop_sequences,
        )

        gen_req = generation_pb2.GenerationRequest(text=text)

        if not self.prefix_id:
            request = generation_pb2.SingleGenerationRequest(
                request=gen_req,
                model_id=self.base_model_name,
                params=params,
            )
        else:
            request = generation_pb2.SingleGenerationRequest(
                request=gen_req,
                model_id=self.base_model_name,
                prefix_id=self.prefix_id,
                params=params,
            )

        # stream GenerationResponse
        try:
            stream_response = self.tgis_client.GenerateStream(
                request, timeout=self.tgis_req_timeout
            )

            for stream_part in stream_response:
                details = TokenStreamDetails(
                    finish_reason=stream_part.stop_reason,
                    generated_tokens=stream_part.generated_token_count,
                    seed=stream_part.seed,
                    input_token_count=stream_part.input_token_count,
                )
                token_list = []
                if stream_part.tokens is not None:
                    for token in stream_part.tokens:
                        token_list.append(
                            GeneratedToken(
                                text=token.text, logprob=token.logprob, rank=token.rank
                            )
                        )
                input_token_list = []
                if stream_part.input_tokens is not None:
                    for token in stream_part.input_tokens:
                        input_token_list.append(
                            GeneratedToken(
                                text=token.text, logprob=token.logprob, rank=token.rank
                            )
                        )
                yield GeneratedTextStreamResult(
                    generated_text=stream_part.text,
                    tokens=token_list,
                    input_tokens=input_token_list,
                    details=details,
                )
        except grpc.RpcError as err:
            raise_caikit_core_exception(err)

    def unary_tokenize(
        self,
        text: str,
    ) -> TokenizationResults:
        """Tokenize unary input using TGIS

        Args:
            text: str
                Text to tokenize
        Returns:
            TokenizationResults
                The token count
        """

        # In case internal client is not configured - tokenization
        # cannot be done (individual modules may already check
        # for this)
        error.value_check(
            "<NLP72786256E>",
            self.tgis_client is not None,
            "Backend must be configured and loaded for tokenization",
        )

        log.debug("Building protobuf request to send to TGIS")

        gen_reqs = [generation_pb2.TokenizeRequest(text=text)]

        request = generation_pb2.BatchedTokenizeRequest(
            requests=gen_reqs,
            model_id=self.base_model_name,
        )

        # Currently, we send a batch request of len(x)==1, so we expect one response back
        with alog.ContextTimer(log.trace, "TGIS request duration: "):
            try:
                batch_response = self.tgis_client.Tokenize(
                    request, timeout=self.tgis_req_timeout
                )
            except grpc.RpcError as err:
                raise_caikit_core_exception(err)

        error.value_check(
            "<NLP38899081E>",
            len(batch_response.responses) == 1,
            f"Got {len(batch_response.responses)} responses for a single request",
        )
        response = batch_response.responses[0]

        return TokenizationResults(
            token_count=response.token_count,
        )



# Standard
from unittest.mock import Mock

# Third Party
from peft import PromptTuningConfig
import pytest

# Local
from caikit_nlp.data_model import TuningConfig
from caikit_nlp.modules.text_generation import TextGeneration
from caikit_nlp.modules.text_generation.peft_config import (
    TuningType,
    get_peft_config,
    resolve_base_model,
)
from caikit_nlp.resources.pretrained_model import HFAutoSeq2SeqLM
from tests.fixtures import (
    SEQ2SEQ_LM_MODEL,
    TINY_MODELS_DIR,
    causal_lm_dummy_model,
    causal_lm_train_kwargs,
    seq2seq_lm_dummy_model,
    seq2seq_lm_train_kwargs,
    temp_config,
)


@pytest.mark.parametrize(
    "train_kwargs,dummy_model",
    [
        (
            "seq2seq_lm_train_kwargs",
            "seq2seq_lm_dummy_model",
        ),
        ("causal_lm_train_kwargs", "causal_lm_dummy_model"),
    ],
)
def test_get_peft_config(train_kwargs, dummy_model, request):
    # Fixtures can't be called directly or passed to mark parametrize;
    # Currently, passing the fixture by name and retrieving it through
    # the request is the 'right' way to do this.
    train_kwargs = request.getfixturevalue(train_kwargs)
    dummy_model = request.getfixturevalue(dummy_model)

    # Define some sample values for testing
    tuning_type = TuningType.PROMPT_TUNING
    tuning_config = TuningConfig(
        num_virtual_tokens=8,
        prompt_tuning_init_method="TEXT",
        prompt_tuning_init_text="Hello world",
    )
    dummy_resource = train_kwargs["base_model"]

    # Call the function being tested
    task_type, output_model_types, peft_config, tuning_type = get_peft_config(
        tuning_type,
        tuning_config,
        dummy_resource,
        dummy_model,
        "float32",
        "{{input}}",
    )

    # Add assertions to validate the behavior of the function
    assert task_type == dummy_resource.TASK_TYPE
    assert output_model_types == dummy_resource.PROMPT_OUTPUT_TYPES
    assert tuning_type == TuningType.PROMPT_TUNING

    # Validation for type & important fields in the peft config
    assert isinstance(peft_config, PromptTuningConfig)
    assert peft_config.num_virtual_tokens == tuning_config.num_virtual_tokens
    assert peft_config.task_type == dummy_resource.TASK_TYPE
    assert peft_config.prompt_tuning_init == tuning_config.prompt_tuning_init_method
    assert peft_config.prompt_tuning_init_text == tuning_config.prompt_tuning_init_text


def test_resolve_model_with_invalid_path_raises():
    """Test passing invalid path to resolve_model function raises"""

    invalid_base_model = "path/../../important"
    with pytest.raises(ValueError):
        resolve_base_model(invalid_base_model, None, "foo")


def test_resolve_model_with_valid_folder_path():
    """Test passing valid folder path to resolve_model function works"""

    model = resolve_base_model(SEQ2SEQ_LM_MODEL, TextGeneration, "float32")

    assert isinstance(model, HFAutoSeq2SeqLM)


def test_resolve_model_works_preloaded_model():

    base_model = HFAutoSeq2SeqLM.bootstrap(SEQ2SEQ_LM_MODEL)
    resolved_model = resolve_base_model(base_model, TextGeneration, "float32")
    assert isinstance(resolved_model, HFAutoSeq2SeqLM)


def test_resolve_model_with_different_base_path_works():

    base_model_name = "T5ForConditionalGeneration"
    with temp_config(base_models_dir=TINY_MODELS_DIR):
        resolved_model = resolve_base_model(base_model_name, TextGeneration, "float32")
        assert isinstance(resolved_model, HFAutoSeq2SeqLM)



"""Tests for sequence classification module
"""
# Standard
import tempfile

# Third Party
from pytest import approx
import pytest

# First Party
from caikit.interfaces.nlp.data_model import ClassificationResult, ClassificationResults

# Local
from caikit_nlp.modules.text_classification import SequenceClassification
from tests.fixtures import SEQ_CLASS_MODEL

## Setup ########################################################################

# Bootstrapped sequence classification model for reusability across tests
# .bootstrap is tested separately in the first test
BOOTSTRAPPED_SEQ_CLASS_MODEL = SequenceClassification.bootstrap(SEQ_CLASS_MODEL)

TEXTS = [
    "The quick brown fox jumps over the lazy dog.",
    "Once upon a time in a land far away",
]

## Tests ########################################################################
# Exact numbers here from the tiny model are not particularly important,
# but we check them here to make sure that the arrays are re-ordered correctly


def test_bootstrap_and_run():
    """Check if we can bootstrap and run sequence classification models"""
    model = SequenceClassification.bootstrap(SEQ_CLASS_MODEL)
    classification_result = model.run(TEXTS[0])
    assert isinstance(classification_result, ClassificationResults)
    assert len(classification_result.results) == 2  # 2 labels

    assert isinstance(classification_result.results[0], ClassificationResult)
    assert classification_result.results[0].label == "LABEL_0"
    assert approx(classification_result.results[0].score) == 0.49526197
    assert classification_result.results[1].label == "LABEL_1"
    assert approx(classification_result.results[1].score) == 0.50473803


def test_bootstrap_and_run_batch():
    """Check if we can bootstrap and run_batch sequence classification models"""
    classification_result_list = BOOTSTRAPPED_SEQ_CLASS_MODEL.run_batch(TEXTS)
    assert len(classification_result_list) == 2

    first_result = classification_result_list[0]
    assert isinstance(first_result, ClassificationResults)
    assert first_result.results[0].label == "LABEL_0"
    assert approx(first_result.results[0].score) == 0.49526197
    assert first_result.results[1].label == "LABEL_1"
    assert classification_result_list[1].results[0].label == "LABEL_0"


def test_load_save_and_run_model():
    """Check if we can load and run a saved model successfully"""
    with tempfile.TemporaryDirectory() as model_dir:
        BOOTSTRAPPED_SEQ_CLASS_MODEL.save(model_dir)
        new_model = SequenceClassification.load(model_dir)
        classification_result = new_model.run(TEXTS[0])
        assert isinstance(classification_result, ClassificationResults)
        assert len(classification_result.results) == 2  # 2 labels

        assert isinstance(classification_result.results[0], ClassificationResult)
        assert classification_result.results[0].label == "LABEL_0"
        assert approx(classification_result.results[0].score) == 0.49526197



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utility functions used for executing run function for text_generation"""

# Standard
from typing import List, Optional, Tuple, Union

# Third Party
from peft.peft_model import PeftModel
from transformers import AutoModel, AutoTokenizer, StoppingCriteria, TextStreamer
import numpy as np
import torch

# First Party
from caikit.core.data_model.producer import ProducerId
from caikit.core.exceptions import error_handler
from caikit.interfaces.nlp.data_model import (
    FinishReason,
    GeneratedTextResult,
    GeneratedTextStreamResult,
    TokenStreamDetails,
)
import alog

# Local
from caikit_nlp.data_model import ExponentialDecayLengthPenalty

log = alog.use_channel("RUN_UTILS")
error = error_handler.get(log)

VALID_DECODING_METHODS = ["GREEDY", "SAMPLING"]

GENERATE_FUNCTION_ARGS = """
    text: str
        Input string to be used to the generation model.
    max_new_tokens: int
        The maximum numbers of tokens to generate.
        Default: 20
    min_new_tokens: int
        The minimum numbers of tokens to generate.
        Default: 0 - means no minimum
    truncate_input_tokens: int
        Truncate inputs to provided number of tokens. This can be
        use to avoid failing due to input being longer than
        configured limits.
        Default: 0 - means don't truncate, thus throw error.
    decoding_method: str
        Parameters for conditionally penalizing / boosting
        candidate tokens during decoding.
        Options: "GREEDY" (default), "SAMPLING"
    top_k: int
        The number of highest probability vocabulary tokens to keep for
        top-k-filtering. Only applicable when decoding_method is SAMPLING.
        Default: 0 - means disabled
    top_p: float
        If set to float < 1, only the smallest set of most probable tokens
        with probabilities that add up to top_p or higher are kept for
        generation. Only applicable when decoding_method is SAMPLING.
        Default: 1.0 - means disabled - 0.0 equivalent to 1.0
    typical_p: float
        Local typicality measures how similar the conditional probability of
        predicting a target token next is to the expected conditional
        probability of predicting a random token next, given the partial text
        already generated. If set to float < 1, the smallest set of the most
        locally typical tokens with probabilities that add up to typical_p
        or higher are kept for generation. Only applicable when decoding_method
        is SAMPLING.
        Default: 1.0 - means disabled - 0.0 equivalent to 1.0
    temperature: float
        The value used to modulate the next token probabilities.
        Only applicable when decoding_method is SAMPLING.
        Default: 1.0 - means disabled - equivalent to 1.0
    repetition_penalty: float
        The more a token is used within generation the more it is penalized
        to not be picked in successive generation passes.
        Default: 1.0 - means no penalty - 0.0 equivalent to 1.0
    max_time: float
        Amount of time in seconds that the query should take maximum.
        NOTE: this does not include network overhead.
        Range: 0-120.0
    exponential_decay_length_penalty: Tuple(int, float)
        This Tuple adds an exponentially increasing length penalty, after
        a certain amount of tokens have been generated. The tuple shall
        consist of: (start_index, decay_factor) where start_index
        indicates where penalty starts and decay_factor represents the factor
        of exponential decay
    stop_sequences: List[str]
        List of strings to be used as stopping criteria
    seed: numpy.uint64
        Random seed to control sampling. Only applicable when decoding_method
        is SAMPLING. Default: None
"""


class Streamer(TextStreamer):
    # The default TextStreamer currently prints to stdout
    # so we override that here
    def on_finalized_text(self, text: str, stream_end: bool = False):
        pass


class SequenceStoppingCriteria(StoppingCriteria):
    # pylint: disable-next=super-init-not-called # false positive: StoppingCriteria is an abc and has no __init__
    def __init__(self, target_sequence_ids):
        self.target_sequence_ids = target_sequence_ids

    def __call__(self, input_ids, scores, **kwargs):
        # Check if the target sequence appears in the generated text
        for seq_id in self.target_sequence_ids:
            if seq_id in input_ids:
                return True  # Stop generation

        return False  # Continue generation

    def __len__(self):
        return 1

    def __iter__(self):
        yield self


def generate_text_func(
    model: "Union[PeftModel, AutoModel]",
    tokenizer: "AutoTokenizer",
    producer_id: ProducerId,
    eos_token: Optional[str],
    text: str,
    max_new_tokens: Optional[int] = 20,
    min_new_tokens: Optional[int] = 0,
    truncate_input_tokens: Optional[int] = 0,
    decoding_method: Optional[str] = "GREEDY",
    top_k: Optional[int] = 0,
    top_p: Optional[float] = 1.0,
    typical_p: Optional[float] = 1.0,
    temperature: Optional[float] = 1.0,
    seed: Optional[np.uint64] = None,
    repetition_penalty: Optional[float] = 1.0,
    max_time: Optional[float] = None,
    exponential_decay_length_penalty: Optional[
        Union[Tuple[int, float], ExponentialDecayLengthPenalty]
    ] = None,
    stop_sequences: Optional[List[str]] = None,
    preserve_input_text: Optional[bool] = True,
    task_type: Optional[str] = None,
    **kwargs,
):
    """
        Args:
            model: PeftModel or transformers.AutoModel
                Peft model or Transformers model
            tokenizer: AutoTokenizer
                Tokenizer to be used with the model
            producer_id: ProducerId
                Caikit producer id associated with the module
            eos_token: str
                End of sequence token to be used with generation
            preserve_input_text: bool
                Applicable only for CAUSAL_LM task type.
                Whether or not the source string should be contained in the generated output,
                e.g., as a prefix. Default True. (Source string will appear as prefix)
            task_type: str or None
                Task type such as CAUSAL_LM, SEQ_2_SEQ_LM, SEQ_CLS or None
            {}
        Returns:
            GeneratedTextResult
    """.format(
        GENERATE_FUNCTION_ARGS
    )

    error.type_check("<NLP85452187E>", str, allow_none=True, eos_token=eos_token)
    error.type_check("<NLP65883534E>", str, text=text)

    error.type_check(
        "<NLP55411551E>",
        int,
        allow_none=True,
        truncate_input_tokens=truncate_input_tokens,
    )

    # NOTE: below is to match TGIS API, where 0 identifies as no truncation
    truncation = truncate_input_tokens != 0

    tok_tensors = tokenizer(
        text,
        truncation=truncation,
        max_length=truncate_input_tokens,
        return_tensors="pt",
    )
    inputs = {k: v.to(model.device) for k, v in tok_tensors.items()}

    input_token_count = tok_tensors["input_ids"].size(1)

    gen_optional_params = __process_gen_args(
        tokenizer,
        max_new_tokens,
        min_new_tokens,
        decoding_method,
        top_k,
        top_p,
        typical_p,
        temperature,
        seed,
        repetition_penalty,
        max_time,
        exponential_decay_length_penalty,
        stop_sequences,
    )

    if "attention_mask" in inputs:
        gen_optional_params["attention_mask"] = inputs["attention_mask"]

    # NOTE: Below is required as `task_id` is a required field for generation
    # with MPT in PEFT. We are manually setting task id to 0 vector since
    # we do not allow setting task specific id anyways.
    if isinstance(model, PeftModel):
        gen_optional_params["task_ids"] = torch.zeros(
            inputs["input_ids"].shape[0], dtype=inputs["input_ids"].dtype
        ).to(model.device)

    with torch.no_grad():
        generate_ids = model.generate(
            input_ids=inputs["input_ids"],
            **gen_optional_params,
            **kwargs,
        )

    token_count = generate_ids.size(1) - 1

    preds = [
        tokenizer.decode(g, skip_special_tokens=True, clean_up_tokenization_spaces=True)
        for g in generate_ids
    ]

    if preserve_input_text is not True:
        generated_text = __postprocess_remove_input_text(
            tokenizer, preds, inputs, task_type
        )
    else:
        generated_text = preds[0]

    if (eos_token and tokenizer.decode(generate_ids[0, -1].item()) == eos_token) or (
        generate_ids[0, -1] == tokenizer.eos_token_id
    ):
        finish_reason = FinishReason.EOS_TOKEN
    elif ("stopping_criteria" in gen_optional_params) and (
        gen_optional_params["stopping_criteria"](
            generate_ids,
            None,  # scores, unused by SequenceStoppingCriteria
        )
    ):
        finish_reason = FinishReason.STOP_SEQUENCE
    else:
        finish_reason = FinishReason.MAX_TOKENS

    return GeneratedTextResult(
        generated_tokens=token_count,
        generated_text=generated_text,
        finish_reason=finish_reason,
        producer_id=producer_id,
        input_token_count=input_token_count,
        seed=seed,
    )


def __postprocess_remove_input_text(tokenizer, preds, inputs, task_type):
    """For Causal LM task types, preserve_input_text set to False will
    remove the input text from generated output.
    """
    if task_type == "CAUSAL_LM":
        prompt_length = len(
            tokenizer.decode(
                inputs["input_ids"][0],
                skip_special_tokens=True,
                clean_up_tokenization_spaces=True,
            )
        )
        generated_text = preds[0][prompt_length:]
    else:
        log.warning(
            "<NLP16125792W>",
            f"preserve_input_text flag is not applicable for task type {task_type}. \
              Returning model generated prediction",
        )
        generated_text = preds[0]
    return generated_text


def generate_text_func_stream(
    model,
    tokenizer,
    producer_id: ProducerId,
    eos_token: str,
    text: str,
    max_new_tokens: Optional[int] = 20,
    min_new_tokens: Optional[int] = 0,
    truncate_input_tokens: Optional[int] = 0,
    decoding_method: Optional[str] = "GREEDY",
    top_k: Optional[int] = 0,
    top_p: Optional[float] = 0.0,
    typical_p: Optional[float] = 0.0,
    temperature: Optional[float] = 1.0,
    seed: Optional[np.uint64] = None,
    repetition_penalty: Optional[float] = 0.0,
    max_time: Optional[float] = None,
    exponential_decay_length_penalty: Optional[
        Union[Tuple[int, float], ExponentialDecayLengthPenalty]
    ] = None,
    stop_sequences: Optional[List[str]] = None,
    **kwargs,
):
    """
        Args:
            model: PeftModel or transformers.AutoModel
                Peft model or Transformers model
            tokenizer: AutoTokenizer
                Tokenizer to be used with the model
            producer_id: ProducerId
                Caikit producer id associated with the module
            eos_token: str
                End of sequence token to be used with generation
            {}
        Returns:
            GeneratedTextResult
    """.format(
        GENERATE_FUNCTION_ARGS
    )
    error.type_check("<NLP53933302E>", str, eos_token=eos_token)
    error.type_check("<NLP61673437E>", str, text=text)

    error.type_check(
        "<NLP60192564E>",
        int,
        allow_none=True,
        truncate_input_tokens=truncate_input_tokens,
    )

    # NOTE: below is to match TGIS API, where 0 identifies as no truncation
    truncation = truncate_input_tokens != 0

    tok_tensors = tokenizer(
        text,
        truncation=truncation,
        max_length=truncate_input_tokens,
        return_tensors="pt",
    )
    inputs = {k: v.to(model.device) for k, v in tok_tensors.items()}

    input_token_count = len(tok_tensors)

    streamer = Streamer(tokenizer)

    gen_optional_params = __process_gen_args(
        tokenizer,
        max_new_tokens,
        min_new_tokens,
        decoding_method,
        top_k,
        top_p,
        typical_p,
        temperature,
        seed,
        repetition_penalty,
        max_time,
        exponential_decay_length_penalty,
        stop_sequences,
    )

    with torch.no_grad():
        # Run tokenized tensors through the rest of the PEFT model
        stream_outputs = model.generate(
            input_ids=inputs["input_ids"],
            attention_mask=inputs["attention_mask"],
            streamer=streamer,
            **gen_optional_params,
            **kwargs,
        )
        details = TokenStreamDetails(
            input_token_count=input_token_count,
            seed=seed,
        )
        for stream_part in stream_outputs:
            gen_text = tokenizer.batch_decode(
                stream_part.detach().cpu().numpy(), skip_special_tokens=True
            )
            yield GeneratedTextStreamResult(
                generated_text=gen_text, details=details, producer_id=producer_id
            )


def __process_gen_args(
    tokenizer,
    max_new_tokens,
    min_new_tokens,
    decoding_method,
    top_k,
    top_p,
    typical_p,
    temperature,
    seed,
    repetition_penalty,
    max_time,
    exponential_decay_length_penalty,
    stop_sequences,
):
    """Utility function to preprocess model generate arguments"""
    error.type_check(
        "<NLP03860680E>", int, allow_none=True, max_new_tokens=max_new_tokens
    )
    error.type_check(
        "<NLP30091276E>", int, allow_none=True, min_new_tokens=min_new_tokens
    )
    error.type_check("<NLP84635843E>", int, allow_none=True, top_k=top_k)
    error.type_check("<NLP55267523E>", float, allow_none=True, top_p=top_p)
    error.type_check(
        "<NLP13670202E>",
        float,
        allow_none=True,
        typical_p=typical_p,
        temperature=temperature,
    )
    error.type_check(
        "<NLP11929418E>",
        float,
        allow_none=True,
        repetition_penalty=repetition_penalty,
        max_time=max_time,
    )
    error.type_check_all(
        "<NLP41311583E>", str, allow_none=True, stop_sequences=stop_sequences
    )
    error.type_check("<NLP28185342E>", int, allow_none=True, seed=seed)

    error.value_check(
        "<NLP80772084E>",
        max_new_tokens >= min_new_tokens,
        "Max new tokens needs to be bigger than min new tokens",
    )

    if isinstance(exponential_decay_length_penalty, ExponentialDecayLengthPenalty):
        exponential_decay_length_penalty = (
            exponential_decay_length_penalty.start_index,
            exponential_decay_length_penalty.decay_factor,
        )

    error.type_check(
        "<NLP81276841E>",
        tuple,
        allow_none=True,
        exponential_decay_length_penalty=exponential_decay_length_penalty,
    )

    error.value_check(
        "<NLP03521360E>",
        decoding_method in VALID_DECODING_METHODS,
        f"Decoding method [{decoding_method}] not in valid decoding methods: "
        f"[{VALID_DECODING_METHODS}]",
    )

    if repetition_penalty == 0.0:
        repetition_penalty = 1.0

    gen_optional_params = {
        "max_new_tokens": max_new_tokens,
        "min_new_tokens": min_new_tokens,
        "repetition_penalty": repetition_penalty,
        "use_cache": True,
        "max_time": max_time,
        "exponential_decay_length_penalty": exponential_decay_length_penalty,
    }

    # TODO: Make decoding parameters enums
    if decoding_method == "SAMPLING":
        gen_optional_params["do_sample"] = True
        gen_optional_params["top_k"] = top_k
        gen_optional_params["top_p"] = top_p
        gen_optional_params["typical_p"] = typical_p
        gen_optional_params["temperature"] = temperature
        gen_optional_params["seed"] = seed

    if stop_sequences and len(stop_sequences) > 0:
        # Tokenize sequences
        stop_sequence_ids = tokenizer.encode(stop_sequences)
        stopping_criteria = SequenceStoppingCriteria(stop_sequence_ids)
        gen_optional_params["stopping_criteria"] = stopping_criteria

    return gen_optional_params



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Standard
from enum import Enum
import os
import re

# Third Party
from peft import MultitaskPromptTuningInit
from transformers import AutoConfig

# First Party
from caikit import get_config
from caikit.core import error_handler
import alog

# Local
from ...data_model import PromptOutputModelType
from ...resources.pretrained_model import PretrainedModelBase
from ...toolkit.data_type_utils import get_torch_dtype
from ...toolkit.verbalizer_utils import is_valid_verbalizer

# NOTE: We do not allow all the methods exposed by MPT / PT, such as `EXACT_SOURCE_TASK`
# since those are for experimental use and would not be useful / applicable
# for end-user use-cases
allowed_tuning_init_methods = [
    "TEXT",
    "RANDOM",
    "ONLY_SOURCE_SHARED",
    "AVERAGE_SOURCE_TASKS",
]

log = alog.use_channel("PFT_CNFG_TLKT")
error = error_handler.get(log)

SOURCE_DIR_VALIDATION_REGEX = re.compile(r"^[-a-zA-Z_0-9\/\.]+")
# 🤮 FIXME: This two dot regex is added as a way to avoid expressions like ..
# giving access to un-intended directories. But this is an ugly hack
# and we need to figure out better solution or better regex
TWO_DOTS_REGEX = re.compile(r"(\.\.)+")


class TuningType(str, Enum):
    PROMPT_TUNING = "PROMPT_TUNING"
    MULTITASK_PROMPT_TUNING = "MULTITASK_PROMPT_TUNING"
    # MULTITASK_PREFIX_TUNING = "MULTITASK_PREFIX_TUNING"
    # P_TUNING = "P_TUNING"
    # PREFIX_TUNING = "PREFIX_TUNING"
    # LORA = "LORA"


def resolve_base_model(base_model, cls, torch_dtype):
    if isinstance(base_model, str):

        error.value_check(
            "<NLP66932773E>",
            re.fullmatch(SOURCE_DIR_VALIDATION_REGEX, base_model)
            and not re.search(TWO_DOTS_REGEX, base_model),
            "invalid characters in base_model name",
        )
        if get_config().base_models_dir:

            base_model_full_path = os.path.join(
                get_config().base_models_dir, base_model
            )
            if os.path.exists(base_model_full_path):
                base_model = base_model_full_path

        model_config = AutoConfig.from_pretrained(
            base_model, local_files_only=not get_config().allow_downloads
        )

        resource_type = None
        for resource in cls.supported_resources:
            if model_config.model_type in resource.SUPPORTED_MODEL_TYPES:
                resource_type = resource
                break

        if not resource_type:
            error(
                "<NLP61784225E>",
                "{} model type is not supported currently!".format(
                    model_config.model_type
                ),
            )
        log.debug("Bootstrapping base resource [%s]", base_model)
        base_model = resource_type.bootstrap(base_model, torch_dtype=torch_dtype)
    return base_model


def get_peft_config(
    tuning_type, tuning_config, base_model, cls, torch_dtype, verbalizer
):

    if tuning_type not in TuningType._member_names_:
        raise NotImplementedError("{} tuning type not supported!".format(tuning_type))

    if tuning_config.prompt_tuning_init_method:
        # NOTE: GK-APR-5-2023
        # MultitaskPromptTuningInit and MultitaskPrefixTuningInit are same at the
        # time of writing, which is a superset of PromptTuningInit
        init_method = tuning_config.prompt_tuning_init_method

        error.value_check(
            "<NLP11848053E>",
            init_method in allowed_tuning_init_methods,
            f"Init method [{init_method}] not in allowed init methods: "
            f"[{allowed_tuning_init_methods}]",
        )

        init_method = MultitaskPromptTuningInit(init_method)
        log.info("Using initialization method [%s]", init_method)

        # If init method provided relates to one that requires source model,
        # make sure the source prompt model is provided.
        if init_method in [
            MultitaskPromptTuningInit.AVERAGE_SOURCE_TASKS,
            MultitaskPromptTuningInit.ONLY_SOURCE_SHARED,
        ]:
            # NOTE: prompt_tuning_init_source_model is currently a path. In future
            # we will replace this with caikit.resources to properly cataloging these
            error.type_check(
                "<NLP89108490E>",
                str,
                prompt_tuning_init_source_model=tuning_config.prompt_tuning_init_source_model,
            )
            tuning_config.prompt_tuning_init_source_model = os.path.join(
                get_config().source_prompt_base,
                tuning_config.prompt_tuning_init_source_model,
            )

            error.file_check(
                "<NLP96030210E>", tuning_config.prompt_tuning_init_source_model
            )
            log.debug(
                "Validated tuning source prompt [%s]",
                tuning_config.prompt_tuning_init_source_model,
            )

    error.type_check("<NLP65714919E>", PretrainedModelBase, base_model=base_model)

    # Validate if tuned output model type is compatible with base model or not
    if not tuning_config.output_model_types:
        output_model_types = base_model.PROMPT_OUTPUT_TYPES
    else:
        # If the first element is not PromptOutputModelType, assume the entire list
        # isn't and convert
        if not isinstance(tuning_config.output_model_types[0], PromptOutputModelType):
            output_model_types = []
            for output_type in tuning_config.output_model_types:
                output_model_types.append(PromptOutputModelType(output_type))
        else:
            output_model_types = tuning_config.output_model_types
        error.value_check(
            "<NLP36947542E>",
            all(
                output_type in base_model.PROMPT_OUTPUT_TYPES
                for output_type in output_model_types
            ),
            "{} not supported for base model type {}".format(
                output_model_types, base_model.MODEL_TYPE
            ),
        )

    error.value_check(
        "<NLP30542004E>",
        len(output_model_types) <= base_model.MAX_NUM_TRANSFORMERS,
        f"Too many output model types. Got {len(output_model_types)}, "
        f"maximum {base_model.MAX_NUM_TRANSFORMERS}",
    )
    # Ensure that our verbalizer is a string and will not render to a hardcoded string
    error.value_check(
        "<NLP83837412E>",
        is_valid_verbalizer(verbalizer),
        "Provided verbalizer is an invalid type or has no renderable placeholders",
    )

    # NOTE: Base model is a resource at this point
    task_type = base_model.TASK_TYPE

    if isinstance(tuning_type, str):
        error.value_check(
            "<NLP65714994E>",
            tuning_type in TuningType._member_names_,
            f"Invalid tuning type [{tuning_type}]. Allowed types: "
            f"[{TuningType._member_names_}]",
        )
        tuning_type = TuningType(tuning_type)
    error.type_check("<NLP65714993E>", TuningType, tuning_type=tuning_type)

    # Coerce the passed model into a resource; if we have one, this is a noop
    # TODO: When splitting up this mono-module, use the configured resource
    #   type of the concrete class to bootstrap
    torch_dtype = get_torch_dtype(torch_dtype)

    # Take tokenizer name/path from the model
    tokenizer_name_or_path = base_model.model.config._name_or_path

    # Build the peft config; this is how we determine that we want a sequence classifier.
    # If we want more types, we will likely need to map this to data model outputs etc.

    # NOTE: We currently only support TEXT as init type, this is to later only easily
    # switch to MPT
    peft_config = cls.create_hf_tuning_config(
        base_model=base_model,
        tuning_type=tuning_type,
        task_type=task_type,
        tokenizer_name_or_path=tokenizer_name_or_path,
        tuning_config=tuning_config,
        output_model_types=output_model_types,
    )

    return task_type, output_model_types, peft_config, tuning_type



"""Tests for text-generation module"""

# Standard
from unittest import mock
import os
import platform
import tempfile

# Third Party
import pytest
import torch

# First Party
from caikit.interfaces.nlp.data_model import GeneratedTextResult
from caikit_tgis_backend import TGISBackend
import caikit

# Local
from caikit_nlp.data_model import ExponentialDecayLengthPenalty, GenerationTrainRecord
from caikit_nlp.modules.text_generation import TextGeneration, TextGenerationTGIS
from caikit_nlp.resources.pretrained_model.hf_auto_seq2seq_lm import HFAutoSeq2SeqLM
from tests.fixtures import set_cpu_device  # noqa
from tests.fixtures import (
    CAUSAL_LM_MODEL,
    SEQ2SEQ_LM_MODEL,
    StubTGISBackend,
    StubTGISClient,
)

SAMPLE_TEXT = "Hello stub"


def test_bootstrap_and_run_causallm():
    """Check if we can bootstrap and run causallm models"""

    model = TextGenerationTGIS.bootstrap(
        CAUSAL_LM_MODEL, load_backend=StubTGISBackend()
    )

    result = model.run(SAMPLE_TEXT, preserve_input_text=True)
    StubTGISClient.validate_unary_generate_response(result)


def test_bootstrap_and_run_seq2seq():
    """Check if we can bootstrap and run seq2seq models"""

    model = TextGenerationTGIS.bootstrap(
        SEQ2SEQ_LM_MODEL, load_backend=StubTGISBackend()
    )

    result = model.run(SAMPLE_TEXT, preserve_input_text=True)
    StubTGISClient.validate_unary_generate_response(result)


def test_bootstrap_and_tokenize_casualllm():
    """Check if we can bootstrap and tokenize text"""
    model = TextGenerationTGIS.bootstrap(
        CAUSAL_LM_MODEL, load_backend=StubTGISBackend()
    )

    result = model.run_tokenizer(SAMPLE_TEXT)
    StubTGISClient.validate_tokenize_response(result)


def test_bootstrap_and_tokenize_seq2seq():
    """Check if we can bootstrap and tokenize text"""
    model = TextGenerationTGIS.bootstrap(
        SEQ2SEQ_LM_MODEL, load_backend=StubTGISBackend()
    )

    result = model.run_tokenizer(SAMPLE_TEXT)
    StubTGISClient.validate_tokenize_response(result)


def test_run_multi_response_errors():
    """Check if multiple responses errors"""
    with mock.patch.object(StubTGISClient, "Generate") as mock_gen_stream:
        fake_response = mock.Mock()
        fake_response.responses = [mock.Mock(), mock.Mock()]
        mock_gen_stream.return_value = fake_response

        model = TextGenerationTGIS.bootstrap(
            SEQ2SEQ_LM_MODEL, load_backend=StubTGISBackend()
        )
        with pytest.raises(ValueError):
            model.run(SAMPLE_TEXT, preserve_input_text=True)


def test_bootstrap_and_save_model():
    """Check if we can bootstrap and save the model successfully"""

    model = TextGenerationTGIS.bootstrap(
        SEQ2SEQ_LM_MODEL, load_backend=StubTGISBackend()
    )

    with tempfile.TemporaryDirectory() as model_dir:
        model.save(model_dir)
        assert os.path.isfile(os.path.join(model_dir, "config.yml"))


def test_save_model_can_run():
    """Check if the model we bootstrap and save is able to load and run successfully"""
    model = TextGenerationTGIS.bootstrap(SEQ2SEQ_LM_MODEL)
    with tempfile.TemporaryDirectory() as model_dir:
        model.save(model_dir)
        del model
        new_model = TextGenerationTGIS.load(
            model_dir, load_backend=StubTGISBackend(mock_remote=True)
        )
        result = new_model.run(SAMPLE_TEXT, preserve_input_text=True)
        StubTGISClient.validate_unary_generate_response(result)


@pytest.mark.skipif(platform.processor() == "arm", reason="ARM training not supported")
def test_local_train_load_tgis(set_cpu_device):
    """Check if the model trained in local module is able to
    be loaded in TGIS module / backend
    """
    train_kwargs = {
        "base_model": HFAutoSeq2SeqLM.bootstrap(
            model_name=SEQ2SEQ_LM_MODEL, tokenizer_name=SEQ2SEQ_LM_MODEL
        ),
        "num_epochs": 1,
        "train_stream": caikit.core.data_model.DataStream.from_iterable(
            [
                GenerationTrainRecord(
                    input="@foo what a cute dog!", output="no complaint"
                )
            ]
        ),
        "torch_dtype": torch.float32,
    }
    model = TextGeneration.train(**train_kwargs)
    with tempfile.TemporaryDirectory() as model_dir:
        model.save(model_dir)
        new_model = TextGenerationTGIS.load(
            model_dir, load_backend=StubTGISBackend(mock_remote=True)
        )
        sample_text = "Hello stub"
        generated_text = new_model.run(sample_text)
        assert isinstance(generated_text, GeneratedTextResult)


def test_remote_tgis_only_model():
    """Make sure that a model can be created and used that will only work with a
    remote TGIS connection (i.e. it has no artifacts)
    """
    model_name = "model-name"
    tgis_backend = StubTGISBackend(mock_remote=True)
    model = TextGenerationTGIS(model_name, tgis_backend=tgis_backend)
    with tempfile.TemporaryDirectory() as model_dir:
        model.save(model_dir)
        TextGenerationTGIS.load(model_dir, load_backend=tgis_backend)


def test_client_lazy_load():
    """
    Test that the TGISBackend client is lazy loaded
    """
    model_name = "model-name"
    tgis_backend = TGISBackend(
        {"connection": {"hostname": "{model_id}.localhost:1234"}}
    )
    model = TextGenerationTGIS(model_name, tgis_backend=tgis_backend)

    # No tgis_backend client and _model_loaded still False
    assert "_client" not in model.__dict__
    assert not model.__dict__.get("_model_loaded", True)

    # Client gets created on accessing ._client
    client = model._client
    assert client is not None
    assert client


### Output streaming tests ##############################################################


def test_bootstrap_and_run_stream_out():
    """Check if we can bootstrap and run_stream_out"""
    model = TextGenerationTGIS.bootstrap(
        SEQ2SEQ_LM_MODEL, load_backend=StubTGISBackend()
    )

    stream_result = model.run_stream_out(SAMPLE_TEXT)
    StubTGISClient.validate_stream_generate_response(stream_result)


def test_run_stream_out_with_runtime_error():
    """Check if runtime error from client raises"""

    with mock.patch.object(StubTGISClient, "GenerateStream") as mock_gen_stream:
        mock_gen_stream.side_effect = RuntimeError("An error!")

        model = TextGenerationTGIS.bootstrap(
            SEQ2SEQ_LM_MODEL, load_backend=StubTGISBackend()
        )
        with pytest.raises(RuntimeError):
            response = model.run_stream_out(SAMPLE_TEXT, preserve_input_text=True)
            # Need to iterate over stream for error
            next(response)


######################## Test run with optional params #####################


def test_bootstrap_and_run_causallm_with_optional_params():
    """Check if we can bootstrap and run causallm models with optional dependencies"""

    model = TextGenerationTGIS.bootstrap(
        CAUSAL_LM_MODEL, load_backend=StubTGISBackend()
    )

    result = model.run(
        SAMPLE_TEXT,
        preserve_input_text=True,
        max_new_tokens=200,
        min_new_tokens=50,
        truncate_input_tokens=10,
        decoding_method="GREEDY",
        repetition_penalty=0.3,
        max_time=10.5,
        exponential_decay_length_penalty=(2, 8),
        stop_sequences=["This is a test"],
    )
    StubTGISClient.validate_unary_generate_response(result)


def test_bootstrap_and_run_stream_out_with_optional_dependencies():
    """Check if we can bootstrap and run_stream_out with optional dependencies"""
    model = TextGenerationTGIS.bootstrap(
        SEQ2SEQ_LM_MODEL, load_backend=StubTGISBackend()
    )

    stream_result = model.run_stream_out(
        SAMPLE_TEXT,
        max_new_tokens=200,
        min_new_tokens=50,
        truncate_input_tokens=10,
        decoding_method="SAMPLING",
        top_k=1,
        top_p=0.1,
        typical_p=0.5,
        temperature=0.75,
        seed=42,
        repetition_penalty=0.3,
        max_time=10.5,
        exponential_decay_length_penalty=ExponentialDecayLengthPenalty(
            start_index=2, decay_factor=1.5
        ),
        stop_sequences=["This is a test"],
    )
    StubTGISClient.validate_stream_generate_response(stream_result)


def test_invalid_optional_params():
    """Check if we an error is thrown when invalid inference params are used to run causallm models"""

    model = TextGenerationTGIS.bootstrap(
        CAUSAL_LM_MODEL, load_backend=StubTGISBackend()
    )

    with pytest.raises(ValueError):
        _ = model.run(
            SAMPLE_TEXT, preserve_input_text=True, max_new_tokens=20, min_new_tokens=50
        )

    with pytest.raises(TypeError):
        _ = model.run(SAMPLE_TEXT, preserve_input_text=True, top_k=0.2)

    with pytest.raises(TypeError):
        _ = model.run(SAMPLE_TEXT, exponential_decay_length_penalty=[2, 2])

    with pytest.raises(ValueError):
        _ = model.run(SAMPLE_TEXT, decoding_method="GREEDY", seed=5)
